import sys
sys.path.append("/home/salmon/workspace/SpDev/SpCommon")
sys.path.append("/home/salmon/workspace/SpDev/SpDB")
from ..util.logger import logger
from spdm.data import connect
import matplotlib.pyplot as pyplot

from spdm.core.plugins.PluginXML import open_xml

if __name__ == '__main__':

    doc = open_xml(["/home/salmon/workspace/SpDev/SpDB/mapping/EAST/imas/3/static"])

    entry = doc.entry

    for coil in entry.pf_active.coil:
        logger.debug(coil.element.geometry.rectangle.__value__())
        break

    # pprint.pprint(collection)
    # a = entry.pf_active.coil[0].element[0].geometry.retangle.r
    # b = a.__fetch__()
    # logger.debug(type(a))
    # logger.debug(type(b))



from __future__ import annotations
import warnings
import typing
from copy import copy, deepcopy
import functools
import collections.abc
import numpy as np

from ..utils.typing import ArrayType
from ..utils.tags import _not_found_
from ..utils.typing import ArrayType, NumericType, array_type, as_array, is_scalar, is_array, numeric_type
from ..utils.logger import logger
from ..utils.misc import group_dict_by_prefix
from ..numlib.interpolate import interpolate

from .Functor import Functor
from .Entry import Entry
from .HTree import HTreeNode, HTree, HTreeNode, List
from .Domain import DomainBase
from .Path import update_tree, Path
from .Functor import Functor, DerivativeOp
from .Expression import Expression

class Piecewise1(Expression):
    """PiecewiseFunction
    A piecewise function. 一维或多维，分段函数
    """

    def __init__(self, piecewise_func: typing.List[typing.Tuple[Expression | float | int, Expression]], **kwargs):
        super().__init__(None, **kwargs)
        self._piecewise = piecewise_func

    def __copy__(self) -> Piecewise:
        res = super().__copy__()
        res._piecewise = self._piecewise
        return res

    def __call__(self, *args, **kwargs) -> NumericType:
        if len(args) == 0:
            return self

        elif any([callable(val) for val in args]):
            return super().__call__(*args, **kwargs)

        elif isinstance(args[0], float):
            for func, cond in self._piecewise:
                if not cond(*args, **kwargs):
                    continue
                else:
                    res = func(*args, **kwargs)
                    break
            else:
                raise RuntimeError(f"Can not fit any condition! {args}")

            return res
        elif isinstance(args[0], array_type):
            res = np.full_like(args[0], np.nan)
            for func, cond in self._piecewise:
                marker = cond(*args, **kwargs)
                if callable(func):
                    _args = [(a[marker] if isinstance(a, array_type) else a) for a in args]
                    _kwargs = {k: (v[marker] if isinstance(v, array_type) else v) for k, v in kwargs.items()}
                    res[marker] = func(*_args, **_kwargs)
                else:
                    res[marker] = func

            return res
        else:
            raise TypeError(f"PiecewiseFunction only support single float or  1D array, {args}")


def piecewise(func_cond, size=None, **kwargs):
    if not isinstance(func_cond, list):
        raise TypeError(f"Illegal type {type(func_cond)}")
    elif all([isinstance(func, (array_type, float, int)) and isinstance(cond, array_type) for func, cond in func_cond]):
        res = np.full_like(func_cond[0][0], np.nan)
        for func, cond in func_cond:
            if np.sum(cond) == 0:
                continue
            if isinstance(func, array_type) and func.size == res.size:
                res[cond] = func[cond]
            else:
                res[cond] = func

        return res
    else:
        return Piecewise(func_cond, **kwargs)




import pathlib
import shutil
import tempfile
import unittest

import h5py
import numpy as np
from spdm.data import File
from spdm.utils.logger import logger

SP_TEST_DATA_DIRECTORY = pathlib.Path("../data")


class TestFileHDF5(unittest.TestCase):
    data = {
        "a": [
            "hello world {name}!",
            "hello world2 {name}!",
        ],
        "b": [1.0, 2, 3, 4],
        "c": "I'm {age}!",
        "d": {
            "e": "{name} is {age}",
            "f": "{address}",
            "g": {"a": 1, "b": 2}
        },
        "h": np.random.random([7, 9])
    }

    def setUp(self) -> None:
        self._temp_dir = tempfile.TemporaryDirectory(prefix="spdm_")
        self.temp_dir = pathlib.Path(self._temp_dir.name)
        shutil.copy(SP_TEST_DATA_DIRECTORY/"test_hdf5_in.h5", self.temp_dir/"test_hdf5_in.h5")

        return super().setUp()

    def tearDown(self) -> None:
        self.temp_dir = None
        self._temp_dir.cleanup()
        del self._temp_dir
        return super().tearDown()

    def test_read(self):
        f_name = self.temp_dir / "test_hdf5_in.h5"

        h5file = h5py.File(f_name)

        first = h5file.keys()

        with File(f_name, mode="r") as f_in:
            second = f_in.read().dump()

    def test_write(self):
        f_name = self.temp_dir / "test_out.h5"
        with File(f_name, mode="w") as f_out:
            f_out.write(self.data)

        self.assertListEqual(list(res.get("a")), self.data["a"])
        self.assertListEqual(list(res.get("b")), self.data["b"])
        self.assertEqual(res.get("d.e"), self.data["d"]["e"])

        self.assertDictEqual(res.get("d"), self.data["d"])

        self.assertTrue(np.array_equal(res.get("h"), self.data["h"]))


if __name__ == '__main__':
    unittest.main()



import os

import matplotlib.pyplot as plt
import numpy as np
from spdm.core.Collection import open_db
from spdm.core.Entry import open_entry
from spdm.utils.logger import logger

os.environ["SP_DATA_MAPPING_PATH"] = "/home/salmon/workspace/fytok_data/mapping"

if __name__ == '__main__':

    # db = open_db("localdb+hdf5:///home/salmon/workspace/output/local_db/{shot}_{run}", mode="rw")
    # entry = open_entry("mdsplus[EAST]://202.127.204.12?tree_name=efit_east#38300")

    entry = open_entry("file+MDSplus[EAST]:///home/salmon/workspace/data/~t/?tree_name=efit_east#38300")

    logger.debug(entry.child("equilibrium/time_slice/100").dump())

    del entry
    # db.insert_one(
    #     {"equilibrium": entry.get(["equilibrium"]).dump()},
    #     shot=123, run=2
    # )



import unittest

import numpy as np
import scipy.constants
from spdm.core.Expression import Expression, Piecewise, Variable
from spdm.utils.logger import logger

TWOPI = scipy.constants.pi * 2.0


class TestExpression(unittest.TestCase):
    def test_variable(self):
        _x = Variable(0, "x")
        _y = Variable(1, "y")

        self.assertEqual(_x.__label__, "x")
        self.assertEqual(str(_x + _y), "x + y")

    def test_expression(self):
        _x = Variable(0, "x")
        _y = Variable(1, "y")
        x = np.linspace(0.0, 1.0, 10)
        y = np.linspace(1.0, 2.0, 10)
        expr = np.sqrt(_x / _y)
        self.assertEqual(expr(1, 2), np.sqrt(0.5))
        self.assertTrue(np.allclose(expr(x, y), np.sqrt(x / y)))

    def test_picewise(self):
        _x = Variable(0)

        r_ped = 0.90  # np.sqrt(0.88)
        Cped = 0.2
        Ccore = 0.4
        chi = Piecewise([_x * 2 * Ccore, Cped], [_x < r_ped, _x >= r_ped])
        self.assertEqual(chi(0.5), (0.5 * 2 * Ccore))
        self.assertEqual(chi(0.95), Cped)

        x = np.linspace(0, 1, 101)

        res = (chi**2)(x)

        self.assertTrue(np.allclose(res[x < r_ped], (x[x < r_ped] * 2 * Ccore) ** 2))
        self.assertTrue(np.allclose(res[x >= r_ped], (Cped) ** 2))

    def test_constant_fun(self):
        value = 1.2345
        xmin = 0.0
        xmax = 1.0
        ymin = 2.0
        ymax = 3.0

        dims_a = np.linspace(xmin, xmax, 5), np.linspace(ymin, ymax, 5)

        fun1 = Function(value)

        xa, ya = np.meshgrid(*dims_a)

        self.assertTrue(np.allclose(fun1(xa, ya), value))

        fun2 = Function(value, *dims_a)

        dims_b = np.linspace(xmin - 0.5, xmax + 0.5, 10), np.linspace(ymin - 0.5, ymax + 0.5, 10)

        xb, yb = np.meshgrid(*dims_b)
        marker = (xb >= xmin) & (xb <= xmax) & (yb >= ymin) & (yb <= ymax)

        e_a = np.full_like(xa, value, dtype=float)

        e_b = np.full_like(xb, value, dtype=float)

        e_b[~marker] = np.nan

        self.assertTrue(np.allclose(fun2(xa, ya), e_a))

        self.assertTrue(np.allclose(fun2(xb, yb), e_b, equal_nan=True))

    def test_delta_fun(self):
        p = 0.5
        value = 1.2345

        fun0 = Function(p, value)

        self.assertTrue(np.isclose(fun0(p), value))

        fun1 = Function([p], [value])

        self.assertTrue(np.isclose(fun1(p), value))

        x = np.linspace(0, 1, 11)

        mark = np.isclose(x, p)

        # logger.debug(fun1(x))
        self.assertTrue(np.allclose(fun1(x)[mark], value))
        self.assertTrue(np.all(np.isnan(fun1(x)[~mark])))

    def test_delta_nd(self):
        p = [0.5, 0.4]
        value = 1.2345

        fun0 = Function(*p, value)

        self.assertTrue(np.isclose(fun0(*p), value))

        dimx = np.linspace(0, 1, 11)
        dimy = np.linspace(0, 1, 11)

        x, y = np.meshgrid(dimx, dimy)

        mark = np.isclose(x, p[0]) & np.isclose(y, p[1])

        self.assertTrue(np.allclose(fun0(x, y)[mark], value))
        self.assertTrue(np.all(np.isnan(fun0(x, y)[~mark])))


if __name__ == "__main__":
    unittest.main()



import matplotlib.pyplot as plt
from fytok.transport.Equilibrium import Equilibrium
from spdm.core.File import File
import numpy as np
from ..util.logger import logger

if __name__ == "__main__":

    eqdsk = File(
        # "/home/salmon/workspace/fytok/examples/data/NF-076026/geqdsk_550s_partbench_case1",
        "/home/salmon/workspace/data/15MA inductive - burn/Standard domain R-Z/High resolution - 257x513/g900003.00230_ITER_15MA_eqdsk16HR.txt",
        # "/home/salmon/workspace/data/Limiter plasmas-7.5MA li=1.1/Limiter plasmas 7.5MA-EQDSK/Limiter_7.5MA_outbord.EQDSK",
        format="geqdsk")

    equilibrium = Equilibrium({"time": 0.0,
                               "time_slice": {
                                   "profiles_1d": eqdsk.entry.pull("profiles_1d"),
                                   "profiles_2d": eqdsk.entry.pull("profiles_2d"),
                                   "coordinate_system": {"Mesh": {"dim1": 100, "dim2": 256}}
                               },
                               "vacuum_toroidal_field":  eqdsk.entry.pull("vacuum_toroidal_field"),
                               })

    o_points, x_points = equilibrium.coordinate_system.critical_points
    psi_axis = o_points[0].psi
    psi_bdry = x_points[0].psi
    psi = np.linspace(psi_axis, psi_bdry, 16)
    # psirz = equilibrium.coordinate_system._psirz
    # data = psirz.__array__()
    # R, Z = psirz.mesh.points
    # logger.debug(type(data))
    contour_set = equilibrium.coordinate_system.find_surface(psi)

    fig = plt.figure()
    # contour_set = plt.contour(R, Z, data, levels=np.linspace(psi_axis, psi_bdry, 16))
    for curv in contour_set:
        if curv is not None:
            plt.plot(*curv.xy.T)

    fig.savefig("/home/salmon/workspace/output/contour.svg", transparent=True)
    # logger.debug(len(contours))



import collections.abc
import typing
from copy import copy, deepcopy

import numpy as np

from .logger import deprecated, logger
from .tags import _not_found_, _undefined_
from .typing import primary_type


class DefaultDict(dict):
    """
    This code creates a dictionary that can have a default value for
    keys that are not yet in the dictionary. It creates a dictionary that inherits
    from the built-in dictionary class. It overrides the __missing__ method to
    return a default value if the key is not in the dictionary. It uses a private
    variable, _factory, that is set to the default factory function. It uses the default
    factory function to generate the default value for the key. It sets the default value
    for the key with the __setitem__ method. It returns the default value.
    """

    def __init__(self, default_factory, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._factory = default_factory

    def __missing__(self, k):
        v = self._factory(k)
        self.__setitem__(k, v)
        return v


_T = typing.TypeVar("_T")


def update_tree(target: _T, key: str | int | list, *args, **kwargs) -> _T:
    pth = None

    if not key and key != 0:
        pass
    elif isinstance(key, str):
        pth = key.split("/")
    elif not isinstance(key, list) and key is not None:
        pth = [key]
    elif len(key) > 0:
        pth = key

    if pth is None:
        if hasattr(target, "_cache"):  # is HTree
            target._cache=update_tree(target._cache, None, *args, **kwargs)

        elif len(args) > 0:
            src = args[0]

            if src is _not_found_:
                pass
            elif target is None or target is _not_found_:
                target = src

            elif target.__class__.__name__ == "HTree":
                target.update(src, **kwargs)

            elif isinstance(target, collections.abc.MutableMapping) and isinstance(src, collections.abc.Mapping):
                # 合并 dict
                for k, v in src.items():
                    update_tree(target, k, v, **kwargs)

            elif kwargs.get("_idempotent", False) is True:
                target = src

            elif isinstance(target, collections.abc.MutableSequence):
                if isinstance(src, collections.abc.Sequence):
                    target.extend(src)
                else:
                    target.append(src)
            elif src is not None:
                target = src

            target = update_tree(target, None, *args[1:], **kwargs)
    else:
        key = pth[0]

        if isinstance(key, str) and key.isdigit():
            key = int(key)

        if isinstance(key, str):
            if target is _not_found_ or target is None:
                target = {}

            # if isinstance(target, collections.abc.MutableMapping):
            if hasattr(target.__class__, key):
                update_tree(getattr(target, key), pth[1:], *args, **kwargs)
            else:
                target[key] = update_tree(target.get(key, _not_found_), pth[1:], *args, **kwargs)

            #     raise RuntimeError(f"Can not update {target} with {key}!")
        else:  # if isinstance(key, int):
            if target is _not_found_ or target is None:
                target = [None] * (key + 1)
            elif isinstance(target, collections.abc.Sequence):
                if key > len(target):
                    target = [*target] + [None] * (key - len(target) + 1)
            # else:
            #     raise TypeError(f"{type(target)} {type(key)}")
            target[key] = update_tree(target[key], pth[1:], *args, **kwargs)

        # else:
        #     raise NotImplementedError(f"{type(key)}")

    return target


 
 

def traversal_tree(d: typing.Any, func: typing.Callable[..., typing.Any]) -> typing.Any:
    if isinstance(d, dict):
        return {k: traversal_tree(v, func) for k, v in d.items()}
    elif isinstance(d, list):
        return [traversal_tree(v, func) for idx, v in enumerate(d)]
    else:
        return func(d)


# def update_tree_recursive(first, second, *args, level=-1, in_place=False, append=False) -> typing.Any:
#     """ 递归合并两个 Hierarchical Tree """
#     if len(args) > 0:
#         return update_tree_recursive(
#             update_tree_recursive(first, second, level=level, in_place=in_place, append=append),
#             *args, level=level, in_place=in_place, append=append)

#     if second is None or second is _not_found_ or level == 0:
#         return first
#     elif first is None or first is _not_found_:
#         return second

#     if in_place:
#         first = copy(first)

#     if isinstance(first, collections.abc.MutableSequence):
#         # 合并 sequence
#         if isinstance(second, collections.abc.Sequence):
#             first.extend(second)
#         else:
#             first.append(second)
#     elif isinstance(first, collections.abc.MutableMapping) and isinstance(second, collections.abc.Mapping):
#         # 合并 dict
#         for k, v in second.items():
#             first[k] = update_tree_recursive(first.get(k, None), v, level=level-1, in_place=in_place)
#     else:
#         first = second
#         # raise TypeError(f"Can not merge {type(first)} with {type(second)}!")

#     return first


class DictTemplate:
    def __init__(self, tmpl, *args, **kwargs):
        self._template = tmpl

    def __missing__(self, key):
        return "{" + key + "}"

    def __getitem__(self, key):
        try:
            res = self._template[key]
        except (KeyError, IndexError):
            raise KeyError(key)

        return res

    def get(self, key, default_value=None):
        try:
            if isinstance(key, str):
                res = _recursive_get(self._template, key.split("."))
            else:
                res = self._template[key]
        except (KeyError, IndexError):
            res = default_value
        return res

    def apply(self, data):
        return tree_apply_recursive(data, lambda s, _template=self: s.format_map(_template), str)[0]


def format_string_recursive(obj, mapping=None):
    class DefaultDict(dict):
        def __missing__(self, key):
            return "{" + key + "}"

    d = DefaultDict(mapping)
    res, _ = tree_apply_recursive(obj, lambda s, _envs=d: s.format_map(d), str)
    return res


def normalize_data(data):
    """Recursively normalizes JSON data.

    Normalizes JSON data by converting all numeric values to floats, and
    converting all strings to unicode. The types argument is a sequence of
    types that should be converted to strings. It defaults to (int, float,
    str).

    Args:
        data: The JSON data to normalize.
        types: A sequence of types that should be converted to strings.

    Returns:
        The normalized JSON data.
    """
    if isinstance(data, primary_type):
        return data
    elif isinstance(data, collections.abc.Mapping):
        return {k: normalize_data(v) for k, v in data.items()}
    elif isinstance(data, collections.abc.Sequence):
        return [normalize_data(v) for v in data]
    elif isinstance(data, collections.abc.Iterable):
        return [normalize_data(v) for v in data]
    else:
        return str(data)


def tree_apply_recursive(obj, op, types=None):
    """Apply op to all elements in a tree-like structure

    Recursively apply a function to all elements in a tree-like structure
    (list, dict, etc.). The function is applied to all elements of the tree.
    It also supports a list of types to include or exclude from the operation.
    Elements of types in the list are included, and elements of types not in
    the list are excluded. If types is None, all elements are included.

    Args:
        obj: The object to be transformed
        op: The function to apply to each element
        types: A list of types to include or exclude from the operation

    Returns:
        The transformed object and a boolean indicating whether a change was made
    """
    changed = False
    is_attr_tree = False
    if types is not None and isinstance(obj, types):
        try:
            res = op(obj)
        except Exception as error:
            res = None
        else:
            changed = True
        return res, changed

    data = obj

    changed = False

    if not isinstance(data, str) and isinstance(data, collections.abc.Sequence):
        for idx, v in enumerate(data):
            new_v, flag = tree_apply_recursive(v, op, types)
            if flag:
                data[idx] = new_v
                changed = True
    elif isinstance(data, collections.abc.Mapping):
        for idx, v in data.items():
            new_v, flag = tree_apply_recursive(v, op, types)
            if flag:
                data[idx] = new_v
                changed = True

    obj = data

    return obj, changed


###############################################################################
# decperecated functions
#
# 下列function皆已经“废弃”，主要功能已经合并进 Path 和 HTree 两个类


@deprecated
def deep_merge_dict(first: dict | list, second: dict, level=-1, in_place=False, force=False) -> dict | list:
    if not in_place:
        first = deepcopy(first)

    if level == 0:
        return first
    elif isinstance(first, collections.abc.Sequence):
        if isinstance(second, collections.abc.Sequence):
            first.extent(second)
        else:
            first.append(second)
    elif isinstance(first, collections.abc.Mapping) and isinstance(second, collections.abc.Mapping):
        for k, v in second.items():
            d = first.get(k, None)
            if isinstance(d, collections.abc.Mapping):
                deep_merge_dict(d, v, level - 1)
            elif d is None:  # or not isinstance(v, collections.abc.Mapping):
                first[k] = v
    elif second is None:
        pass
    else:
        raise TypeError(f"Can not merge dict with {type(second)}!")

    return first


@deprecated
def reduce_dict(d, **kwargs) -> typing.Dict:
    res = {}
    for v in d:
        deep_merge_dict(res, v, in_place=True, **kwargs)
    return res


def _recursive_get(obj, k):
    return obj if len(k) == 0 else _recursive_get(obj[k[0]], k[1:])


@deprecated
def get_value_by_path(data, path, default_value=None):
    # 将路径按 '/' 分割成列表
    if isinstance(path, str):
        segments = path.split("/")
    elif isinstance(path, collections.abc.Sequence):
        segments = path
    else:
        segments = [path]

    # 初始化当前值为 data
    current_value = data
    # 遍历路径中的每一段
    for segment in segments:
        # 如果当前值是一个字典，并且包含该段作为键
        if isinstance(current_value, collections.abc.Mapping) and segment in current_value:
            # 更新当前值为该键对应的值
            current_value = current_value[segment]
        else:
            # 否则尝试将该段转换为整数索引
            try:
                index = int(segment)
                # 如果当前值是一个列表，并且索引在列表范围内
                if isinstance(current_value, list) and 0 <= index < len(current_value):
                    # 更新当前值为列表中对应索引位置的元素
                    current_value = current_value[index]
                else:
                    # 否则返回默认值
                    return default_value
            except ValueError:
                # 如果转换失败，则返回默认值
                return default_value
    # 返回最终的当前值
    return current_value


@deprecated
def set_value_by_path(data, path, value):
    # 将路径按 '/' 分割成列表
    segments = path.split("/")
    # 初始化当前字典为 data
    current_dict = data
    # 遍历路径中除了最后一段以外的每一段
    for segment in segments[:-1]:
        # 如果当前字典包含该段作为键，并且对应的值也是一个字典
        if segment in current_dict and isinstance(current_dict[segment], dict):
            # 更新当前字典为该键对应的子字典
            current_dict = current_dict[segment]
        else:
            # 尝试将该段转换为整数索引
            try:
                index = int(segment)
                # 如果当前字典不包含该段作为键，或者对应的值不是一个列表
                if segment not in current_dict or not isinstance(current_dict[segment], list):
                    # 创建一个空列表作为该键对应的值
                    current_dict[segment] = []
                # 更新当前字典为该键对应的子列表
                current_dict = current_dict[segment]
            except ValueError:
                # 如果转换失败，则抛出一个异常，提示无法继续查找
                raise Exception(f"Cannot find {segment} in {current_dict}")
    # 在当前字典中设置最后一段作为键，给定的值作为值
    last_segment = segments[-1]
    # 尝试将最后一段转换为整数索引
    try:
        index = int(last_segment)
        # 如果当前字典包含最后一段作为键，并且对应的值是一个列表
        if last_segment in current_dict and isinstance(current_dict[last_segment], list):
            # 判断索引是否在列表范围内
            if 0 <= index < len(current_dict[last_segment]):
                # 更新列表中对应索引位置的元素为给定值
                current_dict[last_segment][index] = value
            else:
                # 否则抛出一个异常，提示无法更新列表元素
                raise Exception(f"Index {index} out of range for list {current_dict[last_segment]}")
        else:
            # 否则直接设置最后一段作为键，给定值作为值（此时会创建一个单元素列表）
            current_dict[last_segment] = value
    except ValueError:
        # 如果转换失败，则直接设置最后一段作为键，给定值作为值（此时会覆盖原有列表）
        current_dict[last_segment] = value

    return True


@deprecated
def get_value(*args, **kwargs) -> typing.Any:
    return get_value_by_path(*args, **kwargs)


@deprecated
def get_many_value(
    d: collections.abc.Mapping, name_list: collections.abc.Sequence, default_value=None
) -> collections.abc.Mapping:
    return {k: get_value(d, k, get_value(default_value, idx)) for idx, k in enumerate(name_list)}


@deprecated
def set_value(*args, **kwargs) -> bool:
    return set_value_by_path(*args, **kwargs)


@deprecated
def try_get(obj, path: str, default_value=_undefined_):
    if obj is None or obj is _not_found_:
        return default_value
    elif path is None or path == "":
        return obj

    start = 0
    path = path.strip(".")
    s_len = len(path)
    while start >= 0 and start < s_len:
        pos = path.find(".", start)
        if pos < 0:
            pos = s_len
        next_obj = getattr(obj, path[start:pos], _not_found_)

        if next_obj is not _not_found_:
            obj = next_obj
        elif isinstance(obj, collections.abc.Mapping):
            next_obj = obj.get(path[start:pos], _not_found_)
            if next_obj is not _not_found_:
                obj = next_obj
            else:
                break
        else:
            break

        start = pos + 1
    if start > s_len:
        return obj
    elif default_value is _undefined_:
        raise KeyError(f"Can for find path {path}")
    else:
        return default_value


@deprecated
def try_getattr_r(obj, path: str):
    if path is None or path == "":
        return obj, ""
    start = 0
    path = path.strip(".")
    s_len = len(path)
    while start >= 0 and start < s_len:
        pos = path.find(".", start)
        if pos < 0:
            pos = s_len
        if not hasattr(obj, path[start:pos]):
            break
        obj = getattr(obj, path[start:pos])
        start = pos + 1
    return obj, path[start:]


@deprecated
def getattr_r(obj, path: str):
    # o, p = try_getattr_r(obj, path)

    # if p != '':
    #     raise KeyError(f"Can for find path {path}")
    if type(path) is str:
        path = path.split(".")

    o = obj
    for p in path:
        o = getattr(o, p, None)
        if o is None:
            break
            # raise KeyError(f"Can not get attribute {p} from {o}")
    return o


@deprecated
def getitem(obj, key=None, default_value=None):
    if key is None:
        return obj
    elif hasattr(obj, "__get__"):
        return obj.__get__(key, default_value)
    elif hasattr(obj, "__getitem__"):
        return obj.__getitem__(key) or default_value
    else:
        return default_value


@deprecated
def setitem(obj, key, value):
    if hasattr(obj, "__setitem__"):
        return obj.__setitem__(key, value)
    else:
        raise KeyError(f"Can not setitem {key}")


@deprecated
def iteritems(obj):
    if obj is None:
        return []
    elif isinstance(obj, collections.abc.Sequence):
        return obj
    elif isinstance(obj, collections.abc.Mapping):
        return obj.items()
    else:
        raise TypeError(f"Can not apply 'iteritems' on {type(obj)}!")



import collections.abc
import typing

import numpy as np

from ..utils.typing import ArrayType
from .BBox import BBox
from .GeoObject import GeoObject
from .Line import Segment
from .Point import Point
from .PointSet import PointSet
from .Polyline import Polyline


@GeoObject.register(["polygon", "Polygon"])
class Polygon(PointSet):
    """ Polygon 多边形 """

    def __init__(self, *args,  **kwargs) -> None:
        super().__init__(*args, rank=2, **kwargs)

    @property
    def is_valid(self) -> bool:
        if self.ndim == 2:
            return True
        else:
            raise NotImplementedError(f"{self.__class__.__name__}")

    @property
    def vertices(self) -> typing.Generator[Point, None, None]:
        for p in self._points:
            yield Point(*p)

    @property
    def edges(self) -> typing.Generator[Segment, None, None]:
        pt_iter = self.vertices
        first_pt = next(pt_iter)
        current_pt = first_pt
        while True:
            try:
                next_pt = next(pt_iter)
                yield Segment(current_pt, next_pt)
                current_pt = next_pt
            except StopIteration:
                yield Segment(current_pt, first_pt)
                break

    @property
    def boundary(self) -> Polyline:
        return Polyline(self._points, is_closed=True)


@Polygon.register("rectangle")
class Rectangle(GeoObject):
    def __init__(self, x, y, width, height, ** kwargs) -> None:
        super().__init__(**kwargs)
        self._x = x
        self._y = y
        self._width = width
        self._height = height

    @property
    def bbox(self) -> BBox: return BBox([self._x, self._y], [self._width, self._height])


@Polygon.register("regular_polygon")
class RegularPolygon(Polygon):
    """ Regular Polygon
        正多边形

        cneter: Point or tuple or list  
            中心点
        radius: float
            半径
        num_of_edges: int
            边数
        rot: float
            旋转角度
    """

    def __init__(self, center: Point, radius: float, num_of_edges: int, *args, **kwargs) -> None:
        center = Point(center) if not isinstance(center, Point) else center
        super().__init__(*args, **kwargs)

    @property
    def area(self) -> float:
        return self._impl.area

    @property
    def length(self) -> float:
        return self._impl.length

    @property
    def center(self) -> Point:
        return Point(self._impl.center)

    @property
    def radius(self) -> float:
        return self._impl.radius

    @property
    def inradius(self) -> float:
        return self._impl.inradius

    @property
    def rotation(self) -> float:
        return self._impl.rotation

    @property
    def vertices(self) -> typing.Generator[Point, None, None]:
        for p in self._impl.vertices:
            yield Point(p)



import atexit
import collections.abc
import inspect
import logging
import logging.handlers
import pathlib
import pprint
import sys
from datetime import datetime
from inspect import getframeinfo, stack

from .envs import SP_DEBUG, SP_MPI, SP_LABEL, SP_MPI_RANK, SP_MPI_SIZE

default_formater = logging.Formatter(
    "%(asctime)s [%(name)8s] %(levelname)8s:" "%(pathname)s:%(lineno)d:%(funcName)s: " "%(message)s"
)


MPI_MSG = ""

if SP_MPI is not None and SP_MPI_RANK > 0:
    MPI_MSG = f"[{SP_MPI_RANK}/{SP_MPI_SIZE}]"


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""

    # Black       0;30     Dark Gray     1;30
    # Blue        0;34     Light Blue    1;34
    # Green       0;32     Light Green   1;32
    # Cyan        0;36     Light Cyan    1;36
    # Red         0;31     Light Red     1;31
    # Purple      0;35     Light Purple  1;35
    # Brown       0;33     Yellow        1;33
    # Light Gray  0;37     White         1;37

    grey = "\x1b[0;37m"
    blue = "\x1b[0;34m"
    yellow = "\x1b[1;33m"
    brown = "\x1b[0;33m"
    green = "\x1b[0;32m"
    red = "\x1b[0;31m"
    bold_red = "\x1b[1;31m"
    reset = "\x1b[0m"

    format_normal = (
        "%(asctime)s [%(name)8s] %(levelname)8s:" + MPI_MSG + " %(pathname)s:%(lineno)d:%(funcName)s: %(message)s"
    )

    FORMATS = {
        logging.DEBUG: grey + format_normal + reset,
        logging.INFO: blue + "%(asctime)s [%(name)8s] %(levelname)8s: " + MPI_MSG + "%(message)s" + reset,
        logging.WARNING: brown + format_normal + reset,
        logging.ERROR: red + format_normal + reset,
        logging.CRITICAL: bold_red + format_normal + reset,
    }

    def format(self, record: logging.LogRecord):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        if isinstance(record.msg, collections.abc.Mapping):
            record.msg = pprint.pformat(record.msg)
        return formatter.format(record)


def sp_enable_logging(name, /, handler=None, prefix=None, formater=None):
    m_logger = logging.getLogger(name)

    formater = formater or CustomFormatter()

    if isinstance(handler, str) and handler == "STDOUT":
        handler = logging.StreamHandler(stream=sys.stdout)
    elif handler is None:
        prefix = prefix or os.environ.get("SP_LOG_PREFIX", f"/tmp/sp_log_{os.environ['USER']}/sp_")
        path = pathlib.Path(f"{prefix}{datetime.now().strftime(r'%Y%m%d_%H%M%S')}.log")
        path = path.expanduser().resolve()
        if not path.parent.exists():
            path.parent.mkdir(mode=0o0755, exist_ok=True)
        handler = logging.FileHandler(path)

    if issubclass(type(handler), logging.Handler):
        handler.setFormatter(formater or default_formater)
        handler.setLevel(logging.DEBUG)
        m_logger.addHandler(handler)
    else:
        raise NotImplementedError()

    return m_logger


logger = sp_enable_logging(SP_LABEL, handler="STDOUT")


match SP_DEBUG:
    # case "1" | "true" | "verbose" | "debug" | True:
    #     level = logging.DEBUG
    case "0" | "warning":
        level = logging.WARNING
    case "-2" | "quiet":
        level = logging.CRITICAL
    case "false" | "False" | False:
        level = logging.INFO
    case _:
        level = logging.DEBUG

if SP_MPI_RANK == 0:
    logger.setLevel(level)
else:
    logger.setLevel(logging.ERROR)


def _at_end():
    logger.setLevel(logging.INFO)
    logger.info("The End")
    logging.shutdown()


atexit.register(_at_end)


def deprecated(func):
    """python 修饰器，作用于类方法或函数，当函数或方法被调用时，在标准日志输出 函数或方法定义所在文件和行数"""

    def _wrap(func):
        def wrapped(*args, __fun__=func, **kwargs):
            caller = getframeinfo(stack()[1][0])
            file_name = caller.filename
            line_number = caller.lineno

            if inspect.isfunction(func):
                logger.info(f"Calling deprecated function {file_name}:{line_number}:'{__fun__.__qualname__}' !")
            else:
                logger.info(f"Calling deprecated function {file_name}:{line_number}:{__fun__.__name__}")

            return __fun__(*args, **kwargs)

        return wrapped

    if isinstance(func, str):
        logger.warning(func)
        return lambda o: _wrap(func)
    else:
        return _wrap(func)


def experimental(func):
    def _wrap(func):
        def wrapped(*args, __fun__=func, **kwargs):
            if inspect.isfunction(func):
                logger.warning(f"Experimental function '{__fun__.__qualname__}' !")
                raise DeprecationWarning(__fun__.__qualname__)
            else:
                logger.warning(f"Experimental object {__fun__}")
            return __fun__(*args, **kwargs)

        return wrapped

    if func is None:
        return lambda o: _wrap(func)
    else:
        return _wrap(func)


__all__ = ["logger", "register_logger"]



from __future__ import annotations

import abc
import collections.abc
import functools
import typing
from copy import copy

import numpy as np

from ..utils.logger import logger
from ..utils.typing import ArrayLike, ArrayType, NumericType, array_type, nTupleType
from .BBox import BBox
from .GeoObject import GeoObject
from .Point import Point


class PointSet(GeoObject):
    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 1:
            if isinstance(args[0], collections.abc.Sequence):
                args = args[0]
            elif isinstance(args[0], dict):
                args = [*args[0].values()]

        if len(args) == 0:
            raise RuntimeError(f"{self.__class__.__name__} must have at least one point")
        elif len(args) > 1 and all([isinstance(a, array_type) for a in args]):
            try:
                points = np.stack(args, axis=args[0].ndim)
            except ValueError as error:
                raise ValueError(f"illegal args={args}") from error
        elif isinstance(args[0], PointSet):
            points = args[0]._points
            kwargs = collections.ChainMap(kwargs, args[0]._metadata)
        elif isinstance(args[0], array_type):
            points = args[0]
        else:
            raise TypeError(f"illegal type args[0]={type(args[0])}")

        super().__init__(ndim=points.shape[-1], **kwargs)

        self._points: array_type = points
        self.set_coordinates()

    def set_coordinates(self, *args):
        if len(args) == 0:
            args = self._metadata.get("coordinates", [])

        if len(args) == 0:
            return
        if len(args) == 1 and isinstance(args[0], str):
            args = [x.strip() for x in args[0].split(",")]

        self._metadata["coordinates"] = args

        if isinstance(args, collections.abc.Sequence):
            if len(args) != self.ndim:
                raise ValueError(f"coordinates {args} not match ndim {self.ndim}")

        for idx, coord_name in enumerate(args):
            setattr(self, coord_name, self._points[..., idx])

    def __copy__(self) -> PointSet:
        other: PointSet = super().__copy__()  # type:ignore
        other._points = self._points
        other.set_coordinates()
        return other

    def __array__(self) -> array_type:
        return self._points

    @property
    def points(self) -> typing.List[ArrayType]:
        return [self._points[..., idx] for idx in range(self.ndim)]

    @functools.cached_property
    def bbox(self) -> BBox:
        xmin = [np.min(self._points[..., idx]) for idx in range(self.ndim)]
        dims = [np.max(self._points[..., idx]) - xmin[idx] for idx in range(self.ndim)]
        return BBox(xmin, dims)

    @property
    def boundary(self) -> PointSet | None:
        if self.is_closed:
            return None
        elif self._points.ndim == 2:
            if self.rank == 2:
                return PointSet(self._points)
            elif self.rank == 1:
                return PointSet((self._points[0], self._points[1]))
        else:
            raise NotImplementedError(f"{self.__class__.__name__}.boundary")

    @property
    def vertices(self) -> typing.Generator[Point, None, None]:
        if self._points.ndim != 2:
            raise NotImplementedError(f"{self.__class__.__name__}.vertices for ndim!=2")

        for p in self._points:
            yield Point(*p)



import collections
import collections.abc
import dataclasses
import functools
import importlib
import inspect
import io
import json
import os
import pathlib
import pkgutil
import pwd
import re
import sys
import typing
from enum import Flag, auto
from typing import Sequence, Type
from urllib.parse import ParseResult, urlparse

import numpy as np

from .logger import logger
from .tags import _empty, _not_found_, _undefined_

import re


def camel_to_snake(name):
    # name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
    return name.lower()


def float_unique(d: np.ndarray, x_min=-np.inf, x_max=np.inf) -> np.ndarray:
    if not isinstance(d, np.ndarray):
        d = np.asarray(d)
    d = np.sort(d)
    tag = np.append(True, np.diff(d)) > np.finfo(float).eps * 10
    rtag = np.logical_and(d >= x_min, d <= x_max)
    rtag = np.logical_or(rtag, np.isclose(d, x_min, rtol=1e-8))
    rtag = np.logical_or(rtag, np.isclose(d, x_max, rtol=1e-8))
    tag = np.logical_and(tag, rtag)
    return d[tag]


def array_like(x: np.ndarray, d):
    if isinstance(d, np.ndarray):
        return d
    elif isinstance(d, (int, float)):
        return np.full_like(x, d)
    elif callable(d):
        return np.asarray(d(x))
    else:
        return np.zeros_like(x)
    # else:
    #     raise TypeError(type(d))


# _empty = object()


# class _NOT_FOUND_:
#     def _as_dict(self):
#         return {}

#     def _as_list(self):
#         return []

#     def __serialize__(self):
#         return None


# class _tags(Flag):
#     not_found = auto()
#     undefined = auto()


# _not_found_ = _tags.not_found

# _undefined_ = _tags.undefined


def whoami(obj=None):
    cls = getattr(obj, "__class__", None)
    if cls is not None:
        return f"{cls.__name__}.{inspect.stack()[1][3]}"
    else:
        return inspect.stack()[1][3]


def get_cls_file_path(cls):
    return pathlib.Path(inspect.getfile(cls)).parent


def load_pkg_data(pkgname, path):
    data = pkgutil.get_data(pkgname, path)
    return json.loads(data.decode("utf-8"))


# def try_get(holder, path, default_value=None):
#     path = normalize_path(path)
#     obj = holder

#     for k in path:
#         if isinstance(k, str):
#             next_o = getattr(obj, k, _not_found_)

#             op = getattr(obj.__class__, k, None)
#         else:
#             op = None
#         if op is None:
#             try:
#                 data = obj.__getitem__(k)
#             except Exception:
#                 obj = default_value
#                 break
#             # except IndexError as error:
#             #     raise IndexError(f"{k} > {len(obj)} Error: {error}")
#             else:
#                 obj = data

#         elif isinstance(op, functools.cached_property):
#             obj = op.__get__(obj)
#         elif isinstance(obj, property):
#             obj = op(obj, "fget")(obj)
#         else:
#             obj = default_value

#     return obj


def try_put(holder, path, value):
    res = None
    if isinstance(path, str):
        res = getattr(holder.__class__, path, None)
    if res is None:
        holder.__setitem__(path, value)
    elif isinstance(res, property):
        getattr(res, "fset")(holder, path, value)
    else:
        raise KeyError(f"Can not set attribute {value}!")

    # def try_getitem_r(obj, path):
    #     if path is None or path == '':
    #         return obj, ''
    #     start = 0
    #     path = path.strip(".")
    #     s_len = len(path)
    #     while start >= 0 and start < s_len:
    #         pos = path.find('.', start)
    #         if pos < 0:
    #             pos = s_len
    #         if isinstance(obj, collections.abc.Mapping) and path[start:pos] in obj:
    #             obj = obj.get(path[start: pos])
    #             start = pos+1
    #         else:
    #             break
    #     return obj, path[start:]

    # def getitem_r(obj, path: str):
    #     o, p = try_getitem_r(obj, path)
    #     if p != '':
    #         raise KeyError(f"Can for find path {path}")
    #     return o


def normalize_path(path):
    if path is None:
        path = []
    elif isinstance(path, str):
        path = path.split(".")
    elif not isinstance(path, collections.abc.MutableSequence):
        path = [path]
    return path


def serialize(d):
    if isinstance(d, (int, float, str)):
        return d
    elif hasattr(d, "__array__"):  # numpy.ndarray like
        return d.__array__()
    elif hasattr(d.__class__, "__serialize__"):
        return d.__serialize__()
    elif dataclasses.is_dataclass(d):
        return {f.name: serialize(getattr(d, f.name)) for f in dataclasses.fields(d)}
    elif isinstance(d, collections.abc.Mapping):
        return {k: serialize(v) for k, v in d.items()}
    elif isinstance(d, collections.abc.Sequence):
        return [serialize(v) for v in d]
    else:
        # logger.warning(f"Can not serialize {d.__class__.__name__}!")
        return f"<{d.__class__.__name__}>NOT_SERIALIZABLE!</{d.__class__.__name__}>"
        # raise TypeError(f"Can not serialize {type(d)}!")


def as_file_fun(func=None, *f_args, **f_kwargs):
    """Function wrapper: Convert  first argument (as file path) to File object
    TODO salmon (20190915): specify the position/key of the file path argument
    """

    def _decorate(wrapped):
        @functools.wraps(wrapped)
        def _wrapper(fp, *args, **kwargs):
            if isinstance(fp, str):
                with open(fp, *f_args, **f_kwargs) as fid:
                    res = wrapped(fid, *args, **kwargs)
            elif isinstance(fp, pathlib.Path):
                with fp.open(*f_args, **f_kwargs) as fid:
                    res = wrapped(fid, *args, **kwargs)
            elif isinstance(fp, io.IOBase):
                res = func(fp, *args, **kwargs)
            else:
                raise TypeError(f"Can not convert type({type(fp)}) to file-like object!")

            return res

        return _wrapper

    if func is None:
        return _decorate
    else:
        return _decorate(func)


def as_lazy_object(func, *f_args, **f_kwargs):
    def _decorate(wrapped):
        @functools.wraps(wrapped)
        def _wrapper(*args, **kwargs):
            def fun():
                return wrapped(*args, **kwargs)

            return LazyProxy(fun)

        return _wrapper

    if func is None:
        return _decorate
    else:
        return _decorate(func)


def getlogin():
    try:
        return os.getlogin()
    except Exception:
        return pwd.getpwuid(os.getuid())[0]


def get_username():
    return getlogin()


def _gusses_name(self, name_hint):
    count = sum(1 for k, v in self._graph.nodes.items() if k == name_hint or k.startswith(f"{name_hint}_"))
    if count > 0:
        name_hint = f"{name_hint}_{count}"
    return name_hint


def _try_insert(self, name_hint, node):
    return self.add_node(node, label=self._gusses_name(name_hint or node.__class__.__name__.lower()))


def compile_regex_pattern(pattern):
    try:
        res = re.compile(pattern)
    except re.error:
        res = None
    finally:
        return res


def first_not_empty(*args):
    return next(x for x in args if len(x) > 0)


def guess_class_name(obj):
    if not inspect.isclass(obj):
        cls = obj.__class__
        cls_name = getattr(obj, "__orig_class__", None)
    else:
        cls = obj
        cls_name = None

    return cls_name or f"{cls.__module__}.{cls.__name__}"


def find_duplicate(l: Sequence, atol=1.0e-8):
    for i, val in enumerate(l[:-1]):
        idx = np.flatnonzero(np.abs(l[i + 1 :] - val) < atol)
        if len(idx) > 2:
            yield idx + i


def replace_tokens(value: str, env):
    class _TokenProxy:
        def __init__(self, d):
            if isinstance(d, _TokenProxy):
                self._data = d._data
            else:
                self._data = d

        def __getitem__(self, key):
            value = get_value_by_path(self._data, key.split("."), _not_found_)
            if value is _not_found_:
                return "{" + key + "}"
            else:
                return value

    if isinstance(value, str):
        # 使用 format_map() 方法进行替换，并更新 document 中的 value
        return value.format_map(_TokenProxy(env))
    elif isinstance(value, list):
        return [replace_tokens(v, env) for v in value]
    elif isinstance(value, dict):
        return {k: replace_tokens(v, env) for k, v in value.items()}
    else:
        return value


def fetch_request(url: str) -> typing.Dict:
    """
    根据路径拖回并解析module_file
    """
    if not isinstance(url, ParseResult):
        url = urlparse(url)
    content = None

    if url.scheme in ["http", "https"]:
        # path is a uri
        response = requests.get(url)
        if response.status_code == 200:
            # request is successful
            content = yaml.safe_load(response.text)  # parse the response text as yaml
        else:
            raise FileNotFoundError(url)
    elif url.scheme in ["local", "file", ""]:
        if not os.path.isfile(url.path):
            raise FileNotFoundError(url.path)
        # file exists
        with open(url.path, "r") as file:
            content = yaml.safe_load(file)  # parse the file content as yaml
    else:
        raise NotImplementedError(url)

    return content


primitive_types = (int, bool, str, float, complex, np.ndarray)

builtin_types = (int, bool, str, float, complex, list, dict, set, tuple, np.ndarray)


def group_dict_by_prefix(
    d: collections.abc.Mapping, prefixes: str | typing.List[str], keep_prefix=False, sep="_"
) -> typing.Tuple[typing.Dict, ...]:
    """将 字典 d 根据prefixs进行分组
    prefix 是一个字符串，或者是一个字符串列表
    key具有相同prefix的项会被分到一组， (if keep_prefix is True then prefix会被去掉)
    返回值是一个元组，每元素是分组后的字典，最后一个元素是其他项的字典
    """

    if isinstance(prefixes, str):
        prefixes = prefixes.split(",")

    groups = [None] * len(prefixes) + [{}]

    for key, value in d.items():
        for idx, prefix in enumerate(prefixes):
            if not key.startswith(prefix):
                continue

            if key == prefix:
                if groups[idx] is None:
                    groups[idx] = value
                elif isinstance(value, collections.abc.Mapping):
                    groups[idx].update(value)
                else:
                    raise ValueError(f"prefix {prefix} is not unique")
            else:
                if sep is not None:
                    prefix = prefix + sep

                if not key.startswith(prefix):
                    continue

                if not keep_prefix:
                    key = key[len(prefix) :]

                if groups[idx] is None:
                    groups[idx] = {}
                elif not isinstance(groups[idx], collections.abc.Mapping):
                    raise ValueError(f"prefix {prefix} is not a dict, but {groups[idx]}")

                groups[idx][key] = value
            break
        else:
            groups[-1][key] = value
    return tuple(groups)


def get_positional_argument_count(function):
    """
    获得一个 function 能够接纳的位置参数个数

    Args:
        function: 要获取的函数

    Returns:
        位置参数个数
    """

    import inspect

    signature = inspect.signature(function)
    positional_only_parameters = [
        parameter for parameter in signature.parameters if parameter == inspect.Parameter.POSITIONAL_ONLY
    ]
    return len(positional_only_parameters)


def fun():
    if isinstance(prefixs, str):
        prefixs = [prefixs]

    prefix_dict = {}

    if isinstance(obj, dict):
        prefix_dict = obj
        obj = None

    prefix_len = len(prefix)
    other_dict = {}
    for k, v in d.items():
        if k.startswith(prefix):
            if not keep_prefix:
                k = k[prefix_len:]
            if not isinstance(v, dict) or not isinstance(prefix_dict.get(k, None), dict):
                prefix_dict[k] = v
            else:
                prefix_dict[k] |= v

        elif k != prefix[:-1]:
            other_dict[k] = v

    if obj is None:
        return prefix_dict, other_dict
    else:
        if len(prefix_dict) > 0:
            raise RuntimeWarning(f" {prefix}*  {prefix_dict} are ignored")

        return obj, other_dict





from .Surface import Surface


@Surface.register("bspline_surface")
class BSplineSurface(Surface):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        raise NotImplementedError()



from __future__ import annotations

import typing
import functools
from typing_extensions import Self
import collections
import numpy as np
import numpy.typing as np_tp
from copy import copy, deepcopy

from ..numlib.interpolate import interpolate
from ..utils.logger import logger
from ..utils.tags import _not_found_
from ..utils.typing import ArrayType, NumericType, array_type, get_args, get_origin, as_array
from .Expression import Expression, zero
from .Functor import Functor
from .Path import update_tree, Path
from .Domain import DomainBase
from .HTree import List


class Function(Expression):
    """
    Function

    A function is a mapping between two sets, the _domain_ and the  _value_.
    The _value_  is the set of all possible outputs of the function.
    The _domain_ is the set of all possible inputs  to the function.

    函数定义域为多维空间时，网格采用rectlinear mesh，即每个维度网格表示为一个数组 _dims_ 。
    """

    def __init__(self, *xy: array_type, domain=_not_found_, **kwargs):
        """
        Parameters
        ----------
        *x : typing.Tuple[ArrayType]
            自变量
        y : ArrayType
            变量
        kwargs : 命名参数，
                *           : 用于传递给 Node 的参数
        extrapolate: int |str
            控制当自变量超出定义域后的值
            * if ext=0  or 'extrapolate', return the extrapolated value. 等于 定义域无限
            * if ext=1  or 'nan', return nan
        """

       
        if len(xy) == 1 and isinstance(xy[0], tuple):
            xy = xy[0]

        if len(xy) == 0:
            raise RuntimeError(f"illegal x,y {xy} ")
        elif any([x is _not_found_ for x in xy]):
            raise ValueError(f"Illegal arguments {xy}")

        x = xy[:-1]
        y = xy[-1]

        if domain is _not_found_ and len(x) > 0:
            domain = x
        elif len(x) > 0 and domain is not _not_found_:
            raise RuntimeError(f"Redefined domain {domain} {x}")

        super().__init__(None, domain=domain, **kwargs)

        self._cache = y

    def __copy__(self) -> Self:
        """copy from other"""
        other: Function = super().__copy__()
        other._cache = self._cache
        return other

    def __getitem__(self, idx) -> NumericType:
        return self._cache[idx]

    def __setitem__(self, idx, value) -> None:
        self._ppoly = None
        self._cache[idx] = value

    def __array__(self) -> array_type:
        return self._cache

    def __eval__(self, *args, **kwargs):
        """
        对函数进行编译，用插值函数替代原始表达式，提高运算速度

        NOTE：
            - 由 points，value  生成插值函数，并赋值给 self._ppoly。 插值函数相对原始表达式的优势是速度快，缺点是精度低。
            - 当函数为expression时，调用 value = self.__call__(*points) 。
        TODO:
            - 支持 JIT 编译, support JIT compile
            - 优化缓存
            - 支持多维插值
            - 支持多维求导，自动微分 auto diff

        Parameters
        ----------
        d : typing.Any
            order of function
        force : bool
            if force 强制返回多项式ppoly ，否则 可能返回 Expression or callable
        """
        return self.__ppoly__()(*args, **kwargs)

    def validate(self, value=None, strict=False) -> bool:
        """检查函数的定义域和值是否匹配"""

        m_shape = tuple(self.shape)

        v_shape = ()

        if value is None:
            value = self._cache

        if value is None:
            raise RuntimeError(f" value is None! {self.__str__()}")

        if isinstance(value, array_type):
            v_shape = tuple(value.shape)

        if (v_shape == m_shape) or (v_shape[:-1] == m_shape):
            return True
        elif strict:
            raise RuntimeError(f" value.shape is not match with dims! {v_shape}!={m_shape} ")
        else:
            logger.warning(f" value.shape is not match with dims! {v_shape}!={m_shape} ")
            return False


class Polynomials(Expression):
    """A wrapper for numpy.polynomial
    TODO: imcomplete
    """

    def __init__(
        self,
        coeff,
        *args,
        type: str = None,
        domain=None,
        window=None,
        symbol="x",
        preprocess=None,
        postprocess=None,
        **kwargs,
    ) -> None:
        match type:
            case "chebyshev":
                from numpy.polynomial.chebyshev import Chebyshev

                Op = Chebyshev
            case "hermite":
                from numpy.polynomial.hermite import Hermite

                Op = Hermite
            case "hermite":
                from numpy.polynomial.hermite_e import HermiteE

                Op = HermiteE
            case "laguerre":
                from numpy.polynomial.laguerre import Laguerre

                Op = Laguerre
            case "legendre":
                from numpy.polynomial.legendre import Legendre

                Op = Legendre
            case _:  # "power"
                import numpy.polynomial.polynomial as polynomial

                Op = polynomial

        op = Op(coeff, domain=domain, window=window, symbol=symbol)

        super().__init__(op, *args, **kwargs)
        self._preprocess = preprocess
        self._postprocess = postprocess

    def __eval__(self, x: array_type | float, *args, **kwargs) -> array_type | float:
        if len(args) + len(kwargs) > 0:
            logger.warning(f"Ignore arguments {args} {kwargs}")

        if not isinstance(x, (array_type, float)):
            return super().__call__(x)

        if self._preprocess is not None:
            x = self._preprocess(x)

        y = self._op(x)

        if self._postprocess is not None:
            y = self._postprocess(y)

        return y


def function_like(y: NumericType, *args: NumericType, **kwargs) -> Function:
    if len(args) == 0 and isinstance(y, Function):
        return y
    else:
        return Function(y, *args, **kwargs)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.append("../")

# -- Project information -----------------------------------------------------

project = 'SpDB'
copyright = '2020, 于治 YUZhi@ipp.ac.cn '
author = '于治 (yuzhi@ipp.ac.cn)'

# The full version, including alpha/beta/rc tags
release = '0.0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.napoleon",
              "sphinx.ext.imgmath",
              "sphinx.ext.graphviz",
            #   "sphinx.ext.autodoc",
              "sphinx.ext.todo",
              "sphinx.ext.imgconverter",
              "sphinx.ext.graphviz",
              "sphinx.ext.autosectionlabel",
            #   "sphinxcontrib.bibtex",
              'sphinxcontrib.plantuml',
              "recommonmark",
            #   "docxsphinx"
              ]

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'zh'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
# html_theme = 'alabaster'


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# imgmath_latex = 'xelatex'
imgmath_latex_preamble = r'''
\usepackage{wasysym}
'''
latex_engine = 'xelatex'
latex_elements = {
    'fontpkg': r'''
\setmainfont[Mapping=tex-text]{Noto Serif CJK SC}
\setsansfont[Mapping=tex-text]{Noto Sans Mono CJK SC}
\setmonofont{Noto Sans Mono CJK SC}
''',
    'preamble': r'''
\usepackage[titles]{tocloft}
\cftsetpnumwidth {1.25cm}\cftsetrmarg{1.5cm}
\setlength{\cftchapnumwidth}{0.75cm}
\setlength{\cftsecindent}{\cftchapnumwidth}
\setlength{\cftsecnumwidth}{1.25cm}
\usepackage{polyglossia}
\setdefaultlanguage[variant=american]{english}
\usepackage{wasysym}
\usepackage{esint}
\usepackage{etoolbox}
\patchcmd{\thebibliography}{\section*{\refname}}{}{}{}
''',
    'fncychap': r'\usepackage[Bjornstrup]{fncychap}',
    'printindex': r'\footnotesize\raggedright\printindex',
}
latex_show_urls = 'footnote'

autodoc_member_order = 'bysource'  # "groupwise"

todo_include_todos = True

image_converter_args = ['-verbose']

plantuml_output_format='png'


import typing

import numpy as np
import dataclasses
import collections.abc
import scipy.interpolate
from skimage import measure

from ..core.Field import Field
from ..core.Expression import Variable
from ..geometry.Curve import Curve
from ..geometry.GeoObject import GeoObject
from ..geometry.Point import Point
from ..utils.logger import deprecated, logger
from .optimize import minimize_filter

# import matplotlib.pyplot as plt
# @deprecated
# def find_contours_matplotlib(z: np.ndarray, x: np.ndarray = None, y: np.ndarray = None, /, *args, levels=None, ** kwargs) -> typing.List[typing.List[np.ndarray]]:
#     """
#         args:X: np.ndarray, Y: np.ndarray, Z: np.ndarray
#         TODO: need improvement
#     """
#     fig = plt.figure()
#     contour_set = fig.gca().contour(x, y, z, *args, levels=levels, ** kwargs)
#     return [(contour_set.levels[idx], col.get_segments()) for idx, col in enumerate(contour_set.collections)]


def find_countours_skimage_(
    val: float, z: np.ndarray, x_inter, y_inter
) -> typing.Generator[GeoObject | None, None, None]:
    for c in measure.find_contours(z, val):
        # data = [[x_inter(p[0], p[1], grid=False), y_inter(p[0], p[1], grid=False)] for p in c]
        x = np.asarray(x_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        y = np.asarray(y_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        data = np.stack([x, y], axis=-1)

        if len(data) == 0:
            yield None
        elif data.shape[0] == 1:
            yield Point(*data[0])
        else:
            yield Curve(data)


def find_countours_skimage(vals: list, z: np.ndarray, x: np.ndarray, y: np.ndarray):
    if z.shape == x.shape and z.shape == y.shape:
        pass
    else:
        raise ValueError(f"Array shape does not match! x:{x.shape} , y:{y.shape}, z:{z.shape} ")
    shape = z.shape
    dim0 = np.linspace(0, shape[0] - 1, shape[0])
    dim1 = np.linspace(0, shape[1] - 1, shape[1])
    x_inter = scipy.interpolate.RectBivariateSpline(dim0, dim1, x)
    y_inter = scipy.interpolate.RectBivariateSpline(dim0, dim1, y)

    if not isinstance(vals, (collections.abc.Sequence, np.ndarray)):
        vals = [vals]
    elif isinstance(vals, np.ndarray) and vals.ndim == 0:
        vals = vals.reshape([1])

    for val in vals:
        yield val, find_countours_skimage_(val, z, x_inter, y_inter)

        # count = 0
        # for c in measure.find_contours(z, val):
        #     count += 1
        #     # data = [[x_inter(p[0], p[1], grid=False), y_inter(p[0], p[1], grid=False)] for p in c]
        #     x = np.asarray(x_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        #     y = np.asarray(y_inter(c[:, 0], c[:, 1], grid=False), dtype=float)
        #     data = np.stack([x, y], axis=-1)
        #     if len(data) == 0:
        #         yield val, None
        #     elif data.shape[0] == 1:
        #         yield val, Point(*data[0])
        #     else:
        #         yield val, Curve(data)
        # if count == 0:
        #     yield val, None


def _find_contours(
    *args, values, **kwargs
) -> typing.Generator[typing.Tuple[float, typing.Generator[GeoObject | None, None, None]], None, None]:
    if len(args) == 3:
        z, x, y = args
    elif len(args) == 1:
        if not isinstance(args[0], Field):
            raise TypeError(f"Wrong type of argument! should be Field, got {type(args[0])}")
        f = args[0]
        xy = f.mesh.points
        x = xy[0]
        y = xy[1]
        z = np.asarray(f)
    else:
        raise ValueError(f"Wrong number of arguments! should be 1 or 3, got {len(args)}")

    yield from find_countours_skimage(values, z, x, y, **kwargs)


@dataclasses.dataclass
class OXPoint:
    r: float
    z: float
    value: float


def find_critical_points(psi: Field) -> typing.Tuple[typing.Sequence[OXPoint], typing.Sequence[OXPoint]]:
    opoints = []

    xpoints = []

    R, Z = psi.mesh.points
    _R = Variable(0, "R")
    Bp2 = (psi.pd(0, 1) ** 2 + psi.pd(1, 0) ** 2) / (_R**2)

    D = psi.pd(2, 0) * psi.pd(0, 2) - psi.pd(1, 1) ** 2

    for r, x_z in minimize_filter(Bp2, R, Z):
        p = OXPoint(r, x_z, psi(r, x_z))

        if D(r, x_z) < 0.0:  # saddle/X-point
            xpoints.append(p)
        else:  # extremum/ O-point
            opoints.append(p)

    Rmid, Zmid = psi.mesh.geometry.bbox.origin + psi.mesh.geometry.bbox.dimensions * 0.5

    opoints.sort(key=lambda x: (x.r - Rmid) ** 2 + (x.z - Zmid) ** 2)

    # TODO:

    o_psi = opoints[0].value
    o_r = opoints[0].r
    o_z = opoints[0].z

    # remove illegal x-points . learn from freegs
    # check psi should be monotonic from o-point to x-point

    x_points = []
    s_points = []
    for xp in xpoints:
        length = 20

        psiline = psi(np.linspace(o_r, xp.r, length), np.linspace(o_z, xp.z, length))

        if len(np.unique(psiline[1:] > psiline[:-1])) != 1:
            s_points.append(xp)
        else:
            x_points.append(xp)

    xpoints = x_points

    xpoints.sort(key=lambda x: (x.value - o_psi) ** 2)

    if len(opoints) == 0 or len(xpoints) == 0:
        raise RuntimeError(f"Can not find O-point or X-point! {opoints} {xpoints}")

    return opoints, xpoints


def find_contours(psirz: Field, psi, axis=None) -> typing.Generator[typing.Tuple[float, GeoObject], None, None]:
    """
    if axis is not None:
        only return  closed surface  enclosed axis
        if closed surface does not exists, return None
        number of surface == len(psi)
    else:
        do not guarantee the number of surface == len(psi)
        return all surface ,
    """
    if isinstance(psi, float):
        psi = [psi]

    if axis is None or axis is False:
        for psi_val, surfs in _find_contours(psirz, values=psi):
            for surf in surfs:
                if isinstance(surf, GeoObject):
                    surf.set_coordinates("r", "z")
                yield psi_val, surf

    else:
        # x_point = None
        if isinstance(axis, OXPoint):
            o_point = axis
        elif isinstance(axis, (list, tuple)):
            o_point = OXPoint(*axis)
        else:
            raise TypeError(f"Wrong type of argument! should be OXPoint, list or tuple, got {type(axis)}")

        if isinstance(psi, float):
            psi = [psi]

        current_psi = np.nan
        current_count = 0
        for psi_val, surfs in _find_contours(psirz, values=psi):
            # 累计相同 level 的 surface个数
            # 如果累计的 surface 个数大于1，说明存在磁岛
            # 如果累计的 surface 个数等于0，说明该 level 对应的 surface 不存在
            # 如果累计的 surface 个数等于1，说明该 level 对应的 surface 存在且唯一

            count = 0
            for surf in surfs:
                count += 1
                if surf is None and np.isclose(psi_val, o_point.value):
                    yield psi_val, Point(o_point.r, o_point.z)
                elif isinstance(surf, Point) and all(np.isclose(surf.points, [o_point.r, o_point.z])):
                    yield psi_val, surf  # raise RuntimeError(f"Can not find surface psi={level}")
                elif isinstance(surf, Curve):

                    # theta_0 = np.arctan2(x_point.r-o_point.r, x_point.z-o_point.z)
                    # theta = ((np.arctan2(_R-o_point.r, _Z-o_point.z)-theta_0)+2.0*scipy.constants.pi) % (2.0*scipy.constants.pi)
                    # surf = surf.remesh(theta)
                    surf.set_coordinates("r", "z")
                    yield psi_val, surf
                else:
                    count -= 1
            if count <= 0:
                if np.isclose(psi_val, o_point.value):
                    yield psi_val, Point(o_point.r, o_point.z)
                else:
                    # logger.warning(f"{psi_val} {o_point.psi}")
                    yield psi_val, None
            elif current_count > 1:
                raise RuntimeError(f"Something wrong! Get {current_count} closed surfaces for psi={current_psi}")

            # theta = np.arctan2(surf[:, 0]-o_point.r, surf[:, 1]-o_point.z)
            # logger.debug((max(theta)-min(theta))/(2.0*scipy.constants.pi))
            # if 1.0 - (max(theta)-min(theta))/(2.0*scipy.constants.pi) > 2.0/len(theta):  # open or do not contain o-point
            #     current_count -= 1
            #     continue

            # is_closed = False

            # if np.isclose((theta[0]-theta[-1]) % (2.0*scipy.constants.pi), 0.0):
            #     # 封闭曲线
            #     theta = theta[:-1]
            #     surf = surf[:-1]
            #     # is_closed = True
            # else:  # boundary separatrix
            #     if x_point is None:
            #         raise RuntimeError(f"No X-point ")
            #     # logger.warning(f"The magnetic surface average is not well defined on the separatrix!")
            #     xpt = np.asarray([x_point.r, x_point.z], dtype=float)
            #     b = surf[1:]
            #     a = surf[:-1]
            #     d = b-a
            #     d2 = d[:, 0]**2+d[:, 1]**2
            #     p = xpt-a

            #     c = (p[:, 0]*d[:, 0]+p[:, 1]*d[:, 1])/d2
            #     s = (p[:, 0]*d[:, 1]-p[:, 1]*d[:, 0])/d2
            #     idx = np.flatnonzero(np.logical_and(c >= 0, c**2+s**2 < 1))

            #     if len(idx) == 2:

            #         idx0 = idx[0]
            #         idx1 = idx[1]

            #         theta_x = np.arctan2(xpt[0]-o_point.r, xpt[1]-o_point.z)

            #         surf = np.vstack([[xpt], surf[idx0:idx1]])
            #         theta = np.hstack([theta_x, theta[idx0:idx1]])
            #     else:
            #         raise RuntimeError(f"Can not get closed boundary {o_point}, {x_point} {idx} !")

            # # theta must be strictly increased
            # p_min = np.argmin(theta)
            # p_max = np.argmax(theta)

            # if p_min > 0:
            #     if p_min == p_max+1:
            #         theta = np.roll(theta, -p_min)
            #         surf = np.roll(surf, -p_min, axis=0)
            #     elif p_min == p_max-1:
            #         theta = np.flip(np.roll(theta, -p_min-1))
            #         surf = np.flip(np.roll(surf, -p_min-1, axis=0), axis=0)
            #     else:
            #         raise ValueError(f"Can not convert 'u' to be strictly increased!")
            #     theta = np.hstack([theta, [theta[0]+(2.0*scipy.constants.pi)]])
            #     theta = (theta-theta.min())/(theta.max()-theta.min())
            #     surf = np.vstack([surf, surf[:1]])

            # if surf.shape[0] == 0:
            #     logger.warning(f"{level},{o_point.psi},{(max(theta),min(theta))}")

            # elif surf.shape[0] == 1:
            #     yield level, Point(surf[0][0], surf[0][1])
            # else:
            #     yield level, Curve(surf, theta, is_closed=is_closed)



from __future__ import annotations
import tempfile
import shutil
import pathlib
import os
import typing
import uuid
import contextlib
from ..utils.logger import logger
from ..utils.envs import SP_DEBUG, SP_LABEL
from ..utils.plugin import Pluggable
from ..view import View as sp_view

from .HTree import HTreeNode
from .TimeSeries import TimeSeriesAoS, TimeSlice
from .Edge import InPorts, OutPorts
from .sp_property import SpTree, sp_tree


@sp_tree
class Actor(Pluggable):

    def __init__(self, *args, **kwargs) -> None:
        if self.__class__ is Actor and self.__class__.__dispatch_init__(None, self, *args, **kwargs) is not False:
            return
        SpTree.__init__(self, *args, **kwargs)
        self._uid = uuid.uuid3(uuid.uuid1(clock_seq=0), self.__class__.__name__)

    def _repr_svg_(self):
        return sp_view.display(self.__geometry__(), output="svg")

    def __geometry__(self):
        return None

    @property
    def tag(self) -> str:
        return f"{self._plugin_prefix}{self.__class__.__name__.lower()}"

    @property
    def uid(self) -> uuid.UUID:
        return self._uid

    @contextlib.contextmanager
    def working_dir(self, suffix: str = "", prefix="") -> typing.Generator[pathlib.Path, None, None]:
        pwd = pathlib.Path.cwd()

        working_dir = f"{self.output_dir}/{prefix}{self.tag}{suffix}"

        temp_dir = None

        if SP_DEBUG:
            current_dir = pathlib.Path(working_dir)
            current_dir.mkdir(parents=True, exist_ok=True)
        else:
            temp_dir = tempfile.TemporaryDirectory(prefix=self.tag)
            current_dir = pathlib.Path(temp_dir.name)

        os.chdir(current_dir)

        logger.info(f"Enter directory {current_dir}")

        try:
            yield current_dir
        except Exception as error:
            if temp_dir is not None:
                shutil.copytree(temp_dir.name, working_dir, dirs_exist_ok=True)
            os.chdir(pwd)
            logger.info(f"Enter directory {pwd}")
            logger.exception(f"Failed to execute actor {self.tag}! \n See log in {working_dir} ")
        else:
            if temp_dir is not None:
                temp_dir.cleanup()

            os.chdir(pwd)
            logger.info(f"Enter directory {pwd}")

    @property
    def output_dir(self) -> str:
        return (
            self.get_cache("output_dir", None)
            or os.getenv("SP_OUTPUT_DIR", None)
            or f"{os.getcwd()}/{SP_LABEL.lower()}_output"
        )

    time_slice: TimeSeriesAoS[TimeSlice]
    """ 时间片序列，保存 Actor 历史状态。
        @note: TimeSeriesAoS 长度为 n(=3) 循环队列。当压入序列的 TimeSlice 数量超出 n 时，会调用 TimeSeriesAoS.__full__(first_slice)  
    """

    inports: InPorts
    """ 输入的 Edge，记录对其他 Actor 的依赖。 """

    outports: OutPorts
    """ 输出的 Edge，可视为对于引用（reference）的记录"""

    @property
    def current(self) -> typing.Type[TimeSlice]:
        """当前时间片，指向 Actor 所在时间点的状态。"""
        return self.time_slice.current

    @property
    def previous(self) -> typing.Generator[typing.Type[TimeSlice], None, None]:
        """倒序返回前面的时间片，"""
        yield from self.time_slice.previous

    @property
    def time(self) -> float:
        """当前时间，"""
        return self.time_slice.current.time

    @property
    def iteration(self) -> int:
        """当前时间片执行 refresh 的次数。对于新创建的 TimeSlice，iteration=0"""
        return self.time_slice.current.iteration

    def initialize(self, *args, **kwargs) -> None:
        """初始化 Actor 。"""
        if self.time_slice.is_initializied:
            return

        self.time_slice.initialize(*args, **kwargs)

        tp_hints = typing.get_type_hints(self.__class__.refresh)

        for name, tp in tp_hints.items():
            if name == "return":
                continue
            elif getattr(tp, "_name", None) == "Optional":  # check typing.Optional
                t_args = typing.get_args(tp)
                if len(t_args) == 2 and t_args[1] is type(None):
                    tp = t_args[0]

            self.inports[name].link(type_hint=tp)

        # 查找父节点的输入 ，更新链接 Edge
        self.inports.refresh()
        self.outports.refresh()

    def preprocess(self, *args, **kwargs) -> typing.Type[TimeSlice]:
        """Actor 的预处理，若需要，可以在此处更新 Actor 的状态树。"""

        for k in [*kwargs.keys()]:
            if isinstance(kwargs[k], HTreeNode):
                self.inports[k] = kwargs.pop(k)
        # inputs = {k: kwargs.pop(k) for k in [*kwargs.keys()] if isinstance(kwargs[k], HTreeNode)}

        # self.inports.update(inputs)

        current = self.time_slice.current

        current.refresh(*args, **kwargs)

        return current

    def execute(self, current: typing.Type[TimeSlice], *args) -> typing.Type[TimeSlice]:
        """根据 inports 和 前序 time slice 更新当前time slice"""
        return current

    def postprocess(self, current: typing.Type[TimeSlice]) -> typing.Type[TimeSlice]:
        """Actor 的后处理，若需要，可以在此处更新 Actor 的状态树。
        @param current: 当前时间片
        @param working_dir: 工作目录
        """
        return current

    def refresh(self, *args, **kwargs) -> typing.Type[TimeSlice]:
        """更新当前 Actor 的状态。
        更新当前状态树 （time_slice），并执行 self.iteration+=1

        """

        current = self.preprocess(*args, **kwargs)

        current = self.execute(current, self.time_slice.previous)

        current = self.postprocess(current)

        return current

    def flush(self) -> None:
        """保存当前时间片的状态。
        根据当前 inports 的状态，更新状态并写入 time_slice，
        默认 do nothing， 返回当前时间片
        """
        return self.time_slice.flush()

    def finalize(self) -> None:
        """完成。"""

        self.flush()
        self.time_slice.finalize()

    def fetch(self, *args, **kwargs) -> typing.Type[TimeSlice]:
        return HTreeNode._do_fetch(self.time_slice.current, *args, **kwargs)




from typing import Type

import numpy as np
from spdm.utils.logger import logger


def float_unique(d: np.ndarray, x_min=-np.inf, x_max=np.inf) -> np.ndarray:
    if not isinstance(d, np.ndarray):
        d = np.asarray(d)
    d = np.sort(d)
    tag = np.append(True, np.diff(d)) > np.finfo(float).eps * 10
    rtag = np.logical_and(d >= x_min, d <= x_max)
    rtag = np.logical_or(rtag, np.isclose(d, x_min, rtol=1e-8))
    rtag = np.logical_or(rtag, np.isclose(d, x_max, rtol=1e-8))
    tag = np.logical_and(tag, rtag)
    return d[tag]


def array_like(x: np.ndarray, d):
    if isinstance(d, np.ndarray):
        return d
    elif isinstance(d, (int, float)):
        return np.full_like(x, d)
    elif callable(d):
        return np.asarray(d(x))
    else:
        return np.zeros_like(x)
    # else:
    #     raise TypeError(type(d))


def step_function_approx(x, scale=1.0e-2):
    return np.heaviside(x, scale)
    # return 1 / (1 + np.exp(-x / scale))



from __future__ import annotations

import pathlib
import typing

from ..utils.logger import logger
from ..utils.tags import _undefined_
from .Collection import Collection, InsertOneResult
from .Document import Document
from .Entry import Entry


@Document.register(["directory"])
class Directory(Document):
    """ 
        Default entry for Directory
    """

    def __init__(self, *args, mask=0o777, createparents=False, **kwargs):
        super().__init__(*args, **kwargs)

        self._mask = mask
        self._createparents = createparents

    def __del__(self):
        if self.is_temporary:
            # TODO:
            #   - delete created parents ?
            #   - clear children
            self._dir_path.rmdir

        return super().__del__()

    @property
    def path(self) -> pathlib.Path: return self.url.path

    @property
    def cwd(self) -> pathlib.Path:
        if self.path.is_dir():
            pass
        elif self.is_writable:
            self.path.mkdir(mode=self._mask,
                            parents=self._createparents and self.is_creatable,
                            exist_ok=self.is_creatable)  # ??? logical correct?
        else:
            raise NotADirectoryError(self.path)
        return self.path

    def cd(self, path) -> Directory:
        return self.__class__(self.path/path, mask=self._mask, createparents=self._createparents, mode=self.mode)


@Collection.register(["localdb", "FileCollection"])
class LocalFileDB(Collection):

    def __init__(self, *args, glob: str | None = None, ** kwargs):
        """
        Example:
            file_name="{*}"
        """

        super().__init__(*args, **kwargs)

        if glob is not _undefined_:
            self._glob = glob
        else:
            parts = pathlib.Path(self.url.path).parts

            idx, _ = next(filter(lambda s: '{' in s[1], enumerate(parts)))

            self.url.path = pathlib.Path(*list(parts)[:idx])

            self._glob = "/".join(parts[idx:])

    @property
    def glob(self) -> str: return self._glob

    def guess_id(self, d, auto_inc=True):
        fid = super().guess_id(d, auto_inc=auto_inc)

        if fid is None and auto_inc:
            fid = self.count()

        return fid

    def guess_filepath(self, **kwargs) -> pathlib.Path:
        return self.path/self._glob.format(**kwargs)

    def open_document(self, fid, mode=None) -> Entry:
        fpath = self.guess_filepath({"_id_": fid})
        logger.debug(f"Open Document: {fpath} mode=\"{ mode or self.mode}\"")
        return Document(fpath, mode=mode if mode is not _undefined_ else self.mode).entry

    def insert_one(self, predicate, *args,  **kwargs) -> InsertOneResult:
        doc = self.open_document(self.guess_id(predicate or kwargs, auto_inc=True))
        return doc

    def find_one(self, predicate, projection=None, **kwargs) -> Entry:
        fpath = self.guess_filepath(predicate or kwargs)

        doc = None
        if fpath.exists():
            doc = self.open_document(fpath)
        else:
            for fp in self._path.parent.glob(self._path.name.format(_id="*")):
                if not fp.exists():
                    continue
                doc = self.open_document(fp, mode="r")
                if doc.check(predicate):
                    break
                else:
                    doc = None

        if projection is not None:
            raise NotImplementedError()

        return doc

    def update_one(self, predicate, update,  *args, **kwargs):
        raise NotImplementedError()

    def delete_one(self, predicate,  *args, **kwargs):
        raise NotImplementedError()

    def count(self, predicate=None,   *args, **kwargs) -> int:
        if predicate is None:
            logger.warning("NOT IMPLEMENTED! count by predicate")

        return len(list(self._path.parent.glob(self._path.name.format(_id="*"))))


class CollectionLocalFile(Collection):
    """
        Collection of local files.
    """

    def __init__(self,   *args, file_format=None, mask=None,   **kwargs):
        super().__init__(*args, schema="local", **kwargs)

        logger.debug(self.metadata)

        self._path = pathlib.Path(self.metadata.get("authority", "") +
                                  self.metadata.get("path", ""))  # .replace("*", Collection.ID_TAG)

        self._file_format = file_format

        prefix = self._path

        while "*" in prefix.name or "{" in prefix.name:
            prefix = prefix.parent

        self._prefix = prefix
        self._filename = self._path.relative_to(self._prefix).as_posix()

        if "x" in self._mode:
            self._prefix.mkdir(parents=True, exist_ok=True, mode=mask or 0o777)
        elif not self._prefix.is_dir():
            raise NotADirectoryError(self._prefix)
        elif not self._prefix.exists():
            raise FileNotFoundError(self._prefix)

        self._path = self._path.as_posix().replace("*", "{id:06}")

    @property
    def next_id(self):
        pattern = self._filename if "*" in self._filename else "*"
        return len(list(self._prefix.glob(pattern)))

    def guess_path(self, *args, fid=None, **kwargs):
        return self._path.format(id=fid or self.next_id, **kwargs)

    def find_one(self, *args, projection=None, **kwargs):
        return File(self.guess_path(*args, **kwargs), mode=self.mode).find(projection)

    def insert_one(self, *args, projection=None, **kwargs):
        return File(self.guess_path(*args, **kwargs), mode="x")

    def update_one(self, predicate, update,  *args, **kwargs):
        raise NotImplementedError()

    def delete_one(self, predicate,  *args, **kwargs):
        raise NotImplementedError()



import pprint
import sys
import unittest

import numpy as np
import scipy.integrate
from spdm.utils.RefResolver import RefResolver

sys.path.append("/home/salmon/workspace/SpDev/SpDB")

npoints = 129


class TestRefResolver(unittest.TestCase):

    def test_validate(self):
        resolver = RefResolver(
            base_uri="https://fusionyun.org/schema/draft-00",
            prefetch="pkgdata://spdm/schemas/",
            alias=[["*",   "/home/salmon/workspace/SpDev/SpDB/examples/data/FuYun/modules/*/fy_module.yaml", ],
                   ["*", "/fuyun/modules/*/fy_module.yaml"]])

        logger.debug(resolver.fetch("physics/genray/201213-gompi-2019b"))


if __name__ == '__main__':
    from spdm.utils.logger import logger
    from spdm.utils.RefResolver import RefResolver

    unittest.main()



import unittest
from copy import deepcopy

from spdm.utils.tags import _not_found_
from spdm.data.Schema import Schema
from spdm.data.File import File
import xmlschema

from jsonschema import validate

EQ_PATH = "/home/salmon/workspace/data/15MA inductive - burn/Increased domain R-Z/High resolution - 257x513/g900003.00230_ITER_15MA_eqdsk16VVHR.txt"
DD_PATH = "/home/salmon/workspace/data-dictionary/dd_physics_data_dictionary.xsd"


class TestSchema(unittest.TestCase):

    def test_valid(self):

        xsd_schema = xmlschema.XMLSchema("/home/salmon/workspace/SpDB/python/spdm/data/tests/test_schema.xsd")
        xsd_schema.to_dict()

        # person = {"name": "John Doe", "age": 30}

        # self.assertTrue(xsd_schema.is_valid(person))

    # def test_valid_ids(self):
    #     eq = File(EQ_PATH, format="GEQdsk").read().dump()

    def test_valid_json(self):
        
        from jsonschema import validate
        # Define the schema
        schema = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "number"}
            },
            "required": ["name", "age"]
        }

        # Define the Python object
        data = {"name": "John", "age": 30}

        # Validate the object against the schema
        validate(instance=data, schema=schema)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from scipy import constants
from spdm.data.Expression import Variable
from spdm.data.Function import Function
from spdm.utils.logger import logger
from spdm.data.sp_property import SpTree, sp_property
from spdm.mesh.Mesh import Mesh
from spdm.data.Profile import Profile


class Doo(SpTree):

    @sp_property
    def grid(self) -> Mesh:
        return Mesh()

    psi: np.ndarray = sp_property()
    phi: Profile[float] = sp_property(coordinate1="../psi")


class TestProfile(unittest.TestCase):

    def test_get(self):
        cache = {
            "phi": np.random.rand(128),
            "psi": np.linspace(0, 1, 128)
        }
        doo = Doo(cache)

        doo.phi(np.linspace(0, 1.0, 256))

        self.assertTrue(np.all(np.isclose(doo.phi.__array__(), cache["phi"])))
        self.assertTrue(np.all(np.isclose(doo.phi.points[0], cache["psi"])))

    def test_prop_expr(self):
        cache = {
            "phi": Variable(0, "psi")*2,
            "psi": np.linspace(0, 1, 128)
        }

        doo = Doo(cache)

        doo.phi(np.linspace(0, 1.0, 256))

        self.assertTrue(np.allclose(doo.phi.__array__(), cache["psi"]*2.0))
        self.assertTrue(np.allclose(doo.phi.points[0], cache["psi"]))


if __name__ == '__main__':
    unittest.main()



import os

# from spdm.utils.logger import logger

ENABLE_JAX = False  # os.environ.get("SP_ENABLE_JAX", False)

if ENABLE_JAX:
    from jax import numpy
    from jax import scipy
    # from jax.scipy.optimize import minimize
    # logger.info(f"Using JAX \t: {jax.__version__}")
else:
    import numpy
    import scipy
    # import scipy.constants as constants
    # import scipy.interpolate as interpolate
    # from scipy.optimize import fsolve, minimize, root_scalar

#     logger.info(f"Using SciPy \t: {scipy.__version__}")
#     logger.info(f"Using NumPy \t: {np.version.full_version}")



from __future__ import annotations
from typing import Generator
from typing_extensions import Self
import collections.abc
import abc
import pathlib
import typing
import types
from copy import copy, deepcopy


from ..utils.misc import get_positional_argument_count
from ..utils.logger import deprecated, logger
from ..utils.tags import _not_found_, _undefined_
from ..utils.typing import (
    ArrayType,
    NumericType,
    array_type,
    as_array,
    get_args,
    get_origin,
    isinstance_generic,
    numeric_type,
    serialize,
    get_type,
    type_convert,
)

from .Entry import Entry, open_entry
from .Path import Path, PathLike, Query, as_path, update_tree, merge_tree, path_like


class HTreeNode:
    @classmethod
    def _parser_args(cls, *args, _parent=None, _entry=None, **kwargs):
        if len(args) == 0:
            _cache = _not_found_
        elif isinstance(args[0], (collections.abc.MutableMapping, collections.abc.MutableSequence)):
            _cache = args[0]
            args = args[1:]
        elif isinstance(args[0], (str, Entry)):
            _cache = {"$entry": args[0]}
            args = args[1:]
        else:
            _cache = _not_found_

        if _entry is None:
            _entry = []
        elif not isinstance(_entry, collections.abc.Sequence) or isinstance(_entry, str):
            _entry = [_entry]

        _entry = list(args) + _entry

        metadata = deepcopy(getattr(cls, "_metadata", {}))

        if isinstance(_cache, dict):
            t_entry = _cache.pop("$entry", _not_found_)
            if t_entry is _not_found_:
                pass
            elif t_entry is False:
                _entry = None
            elif not isinstance(t_entry, list):
                _entry = [t_entry] + _entry
            else:
                _entry = t_entry + _entry

            # for k in list(_cache.keys()):
            #     if k.startswith("@"):
            #         metadata[k[1:]] = _cache.pop(k)

        metadata.update(kwargs)

        if not isinstance(_entry, list):
            raise RuntimeError(f"Can not parser _entry {_entry}")

        _entry = open_entry(
            sum([e if isinstance(e, list) else [e] for e in _entry if e is not None and e is not _not_found_], [])
        )

        default_value = metadata.pop("default_value", _not_found_)

        if default_value is not _not_found_:
            _cache = update_tree(deepcopy(default_value), _cache)

        return _cache, _entry, _parent, metadata

    def __init__(self, *args, **kwargs) -> None:
        self._cache, self._entry, self._parent, self._metadata = self.__class__._parser_args(*args, **kwargs)

    def __copy__(self) -> Self:
        if isinstance(self._cache, dict):
            cache = {k: copy(value) for k, value in self._cache.items()}
        elif isinstance(self._cache, list):
            cache = [copy(value) for k, value in self._cache.items()]
        else:
            cache = deepcopy(self._cache)

        res = self.__class__(cache, _entry=self._entry)

        res._metadata = deepcopy(self._metadata)

        return res

    @classmethod
    def _do_serialize(cls, source: typing.Any, dumper: Entry | typing.Callable[..., typing.Any] | bool) -> _T:
        if source is _not_found_:
            return source if not isinstance(dumper, Entry) else dumper

        elif hasattr(source.__class__, "__serialize__"):
            return source.__serialize__(dumper)

        elif isinstance(source, dict):
            if isinstance(dumper, Entry):
                for k, v in source.items():
                    cls._do_serialize(v, dumper.child(k))
                res = dumper
            else:
                res = {k: cls._do_serialize(v, dumper) for k, v in source.items()}

        elif isinstance(source, list):
            if isinstance(dumper, Entry):
                for k, v in enumerate(source):
                    cls._do_serialize(v, dumper.child(k))
                res = dumper
            else:
                res = [cls._do_serialize(v, dumper) for v in source]

        elif isinstance(dumper, Entry):
            dumper.insert(source)
            res = dumper

        elif callable(dumper):
            res = dumper(source)

        elif dumper is True:
            res = deepcopy(source)

        else:
            res = source

        return res

    def __serialize__(self, dumper: Entry | typing.Callable[..., typing.Any] | bool = True) -> Entry | typing.Any:
        """若 dumper 为 Entry，将数据写入 Entry
           若 dumper 为 callable，将数据传入 callable 并返回结果
           若 dumper 为 True，返回数据的拷贝
           否则返回序列化后的数据

        Args:
            target (Entry, optional): 目标入口. Defaults to None.
            copier (typing.Callable[[typing.Any], typing.Any] | bool, optional): copier 拷贝器，当 target 为 None 有效。若为 True 则通过 copy 函数返回数据的拷贝. Defaults to None.

        Returns:
            typing.Any: 若 target is None，返回原始数据，否则返回 target
        """
        if self._cache is _not_found_:
            if self._entry is not None:
                return self._entry.dump(dumper)
            else:
                return self._do_serialize(self.__value__, dumper)
        else:
            return self._do_serialize(self._cache, dumper)

    @classmethod
    def __deserialize__(cls, *args, **kwargs) -> typing.Type[HTree]:
        return cls(*args, **kwargs)

    def __duplicate__(self, *args, _parent=_not_found_, **kwargs):
        if _parent is _not_found_:
            _parent = self._parent

        cls = get_type(self)

        if len(args) == 0:
            args = [deepcopy(self._cache)]

        return cls(
            *args,
            _parent=_parent,
            _entry=self._entry,
            **collections.ChainMap(kwargs, deepcopy(self._metadata)),
        )

    @property
    def __name__(self) -> str:
        return self._metadata.get("name", self.__class__.__name__)

    @property
    def __path__(self) -> typing.List[str | int]:
        if isinstance(self._parent, HTreeNode):
            return self._parent.__path__ + [self.__name__]
        else:
            return [self.__name__]

    @property
    def __label__(self) -> str:
        return self._metadata.get("label", None) or ".".join(self.__path__)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} name='{'.'.join(self.__path__)}' />"

    def __repr__(self) -> str:
        return ".".join(self.__path__)

    @property
    def __root__(self) -> HTree | None:
        p = self
        # FIXME: ids_properties is a work around for IMAS dd until we found better solution
        while isinstance(p, HTreeNode) and getattr(p, "ids_properties", None) is None:
            p = p._parent
        return p

    @property
    def __value__(self) -> typing.Any:
        return self._cache

    def __array__(self) -> ArrayType:  # for numpy
        return as_array(self.__value__)

    def __null__(self) -> bool:
        """判断节点是否为空，若节点为空，返回 True，否则返回 False
        @NOTE 长度为零的 list 或 dict 不是空节点
        """
        return self._cache is None and self._entry is None

    def __empty__(self) -> bool:
        return (self._cache is _not_found_ or len(self._cache) == 0) and (self._entry is None)

    def __bool__(self) -> bool:
        return self.__null__() or self.__empty__() or bool(self.__value__)

    def __equal__(self, other) -> bool:
        if isinstance(other, HTreeNode):
            return other.__equal__(self._cache)
        else:
            return self._cache == other

    @staticmethod
    def _do_fetch(obj, *args, **kwargs):
        if hasattr(obj.__class__, "fetch"):
            return obj.fetch(*args, **kwargs)

        elif isinstance(obj, dict):
            return {k: HTreeNode._do_fetch(v, *args, **kwargs) for k, v in obj.items()}

        elif isinstance(obj, (list, set, tuple)):
            return obj.__class__([HTreeNode._do_fetch(v, *args, **kwargs) for v in obj])

        elif callable(obj):
            return obj(*args, **kwargs)

        else:
            return deepcopy(obj)

            # raise RuntimeError(f"Unknown argument {obj} {args} {kwargs}")

    def fetch(self, *args, **kwargs) -> Self:
        return self.__duplicate__(HTreeNode._do_fetch(self._cache, *args, **kwargs))

    def flush(self):
        """TODO: 将 cache 内容 push 入 entry"""
        return NotImplemented


null_node = HTreeNode(_not_found_)


_T = typing.TypeVar("_T")


class HTree(HTreeNode, typing.Generic[_T]):
    """Hierarchical Tree:
    - 其成员类型为 _T，用于存储一组数据或对象，如列表，字典等

    - 一种层次化的数据结构，它具有以下特性：
    - 树节点也可以是列表 list，也可以是字典 dict
    - 叶节点可以是标量或数组 array_type，或其他 type_hint 类型
    - 节点可以有缓存（cache)
    - 节点可以有父节点（_parent)
    - 节点可以有元数据（metadata), 包含： 唯一标识（id), 名称（name), 单位（units), 描述（description), 标签（tags), 注释（comment)
    - 任意节点都可以通过路径访问
    - `get` 返回的类型由 `type_hint` 决定，默认为 Node



    """

    def __contains__(self, path) -> bool:
        return bool(self.get(path, Path.tags.exists)) is True

    #############################################
    # 对后辈节点操作，支持路径

    def put(self, path, value, *args, **kwargs):
        return as_path(path).update(self, value, *args, **kwargs)

    def get(self, path: PathLike, default_value: _T = _not_found_, *args, **kwargs) -> _T:
        return as_path(path).find(self, *args, default_value=default_value, **kwargs)

    @typing.final
    def pop(self, path, default_value=_not_found_):
        path = as_path(path)

        value = path.get(self, _not_found_)

        if value is not _not_found_:
            path.put(self, _not_found_)
            return value
        else:
            return default_value

    @typing.final
    def __getitem__(self, path) -> _T:
        return self.get(path, default_value=_undefined_)

    @typing.final
    def __setitem__(self, path, value) -> None:
        self.put(path, value, Path.tags.overwrite)

    @typing.final
    def __delitem__(self, path) -> None:
        return self.put(path, _not_found_, _idempotent=True)

    @typing.final
    def get_cache(self, path, default_value: _T = _not_found_) -> _T:
        path = as_path(path)
        res = path.get(self._cache, _not_found_)

        if res is _not_found_ and self._entry is not None:
            res = self._entry.get(path, _not_found_)

        if res is _not_found_:
            res = default_value

        return res

    #############################################
    # 当子节点操作，不支持路径
    @abc.abstractclassmethod
    def children(self) -> typing.Generator[_T, None, None]:
        """alias of for_each"""
        return

    @abc.abstractclassmethod
    def __iter__(self) -> typing.Generator[_T | str, None, None]:
        """遍历 children"""
        return

    @typing.final
    def __len__(self) -> int:
        return int(self._find_(None, Path.tags.count) or 0)

    @typing.final
    def __contains__(self, key) -> bool:
        return bool(self.find(key, Path.tags.exists) is True)

    @typing.final
    def __equal__(self, other) -> bool:
        return bool(self._find_(None, Path.tags.equal, other))

    ###############################################################################
    # RESTful API

    @typing.final
    def insert(self, *args, **kwargs) -> None:
        return self._insert_(*args, **kwargs)

    @typing.final
    def update(self, *args, **kwargs) -> None:
        if len(args) == 0 or isinstance(args[0], dict):
            args = [None, *args]
        return self._update_(*args, **kwargs)

    @typing.final
    def remove(self, *args, **kwargs) -> bool:
        return self._remove_(*args, **kwargs)

    @typing.final
    def find(self, *args, **kwargs) -> _T:
        return self._find_(*args, **kwargs)

    @typing.final
    def find_cache(self, path, *args, default_value=_not_found_, **kwargs) -> typing.Any:
        res = Path._do_find(self._cache, path, *args, default_value=_not_found_, **kwargs)
        if res is _not_found_ and self._entry is not None:
            res = self._entry.child(path).find(*args, default_value=_not_found_, **kwargs)
        if res is _not_found_:
            res = default_value
        return res

    @typing.final
    def get_cache(self, path, default_value=_not_found_) -> typing.Any:
        return self.find_cache(path, default_value=default_value)

    @typing.final
    def for_each(self, *args, **kwargs) -> typing.Generator[typing.Tuple[int | str, _T], None, None]:
        yield from self._for_each_(*args, **kwargs)

    # -----------------------------------------------------------------------------
    # 内部接口

    def _insert_(self, value: typing.Any, _op=None, *args, **kwargs):
        return self._update_(None, value, _op or Path.tags.insert, *args, **kwargs)

    def _remove_(self, key: str | int, *args, _deleter: typing.Callable = None, **kwargs) -> bool:
        """删除节点：
        - 将 _cahce 中 path 对应的节点设置为 None，这样访问时不会触发 _entry
        - 若 path 为 None，删除所有子节点
        """
        if callable(_deleter):
            return _deleter(self, key)
        else:
            return self._update_(key, _not_found_, _op=Path.tags.remove, *args, **kwargs)

    def _update_(self, key: str | int, value: typing.Any, _op: Path.tags = None, *args, _setter=None, **kwargs):
        if (key is None or key == []) and value is self:
            pass

        elif isinstance(key, str) and key.startswith("@"):
            value = Path._do_update(self._metadata, key[1:], value, _op, *args, **kwargs)
            if value is not _not_found_:
                return value

        elif callable(_setter):
            _setter(self, key, value)

        else:
            self._cache = Path._do_update(self._cache, key, value, _op, *args, **kwargs)

        return self

    def __missing__(self, key: str | int) -> typing.Any:
        raise KeyError(f"{self.__class__.__name__}.{key} is not assigned! ")

    def _find_(self, key, *args, _getter=None, default_value=_undefined_, **kwargs) -> _T:
        """获取子节点/或属性
        搜索子节点的优先级  cache > getter > entry > default_value
        当 default_value 为 _undefined_ 时，若 cache 中找不到节点，则从 entry 中获得

        """

        if isinstance(key, str) and key.startswith("@"):
            value = Path._do_find(self._metadata, key[1:], *args, default_value=_not_found_)
            if value is not _not_found_:
                return value

        if len(args) > 0:
            value = Path._do_find(self._cache, key, *args, default_value=_not_found_, **kwargs)
            if value is _not_found_:
                if self._entry is not None:
                    value = self._entry.child([key]).find(*args, default_value=default_value, **kwargs)
                else:
                    value = default_value

        else:
            if isinstance(key, int):
                if key < len(self._cache):
                    value = self._cache[key]
                else:
                    value = _not_found_
            else:
                value = Path._do_find(self._cache, key, default_value=_not_found_)

            # if isinstance(default_value, dict):
            #     value = update_tree(deepcopy(default_value), value)
            #     default_value = _not_found_

            _entry = self._entry.child(key) if self._entry is not None else None

            if value is _not_found_ and callable(_getter):
                if get_positional_argument_count(_getter) == 2:
                    value = _getter(self, key)
                else:
                    value = _getter(self)

            if value is _not_found_ and _entry is not None and default_value is _undefined_:
                value = _entry.get(default_value=_not_found_)
                _entry = None

            if value is _not_found_ and _entry is None:
                value = default_value
                default_value = _not_found_

            if value is _undefined_:
                value = self.__missing__(key)

            value = self._type_convert(value, key, _entry=_entry, default_value=default_value, **kwargs)

            if key is None and isinstance(self._cache, collections.abc.MutableSequence):
                self._cache.append(value)
            elif isinstance(key, str) and isinstance(self._cache, collections.abc.MutableSequence):
                self._cache = Path._do_update(self._cache, key, value)
            elif isinstance(key, int) and key <= len(self._cache):
                self._cache.extend([_not_found_] * (key - len(self._cache) + 1))
                self._cache[key] = value
            else:
                self._cache[key] = value

        return value

    def _for_each_(self, *args, **kwargs) -> typing.Generator[typing.Tuple[int | str, _T], None, None]:
        if (self._cache is _not_found_ or len(self._cache) == 0) and self._entry is not None:
            for k, v in self._entry.for_each(*args, **kwargs):
                if not isinstance(v, Entry):
                    yield k, self._type_convert(v, k)
                else:
                    yield k, self._type_convert(_not_found_, k, _entry=v)

        elif self._cache is not None:
            for k, v in Path._do_for_each(self._cache, [], *args, **kwargs):
                _entry = self._entry.child(k) if self._entry is not None else None
                v = self._type_convert(v, k, _entry=_entry)
                yield k, v

    ################################################################################
    # Private methods

    def _type_hint_(self, key: str | int = None) -> typing.Type:
        """当 key 为 None 时，获取泛型参数，若非泛型类型，返回 None，
        当 key 为字符串时，获得属性 property 的 type_hint
        """

        if isinstance(key, str) and key.startswith("@"):
            return None

        tp = None

        if isinstance(key, str):
            cls = getattr(self, "__orig_class__", self.__class__)

            tp = typing.get_type_hints(get_origin(cls)).get(key, None)

        if tp is None:
            tp = get_args(getattr(self, "__orig_class__", None) or self.__class__)
            tp = tp[-1] if len(tp) > 0 else None

        return tp

    def _type_convert(
        self,
        value: typing.Any,
        _key: int | str,
        default_value: typing.Any = _not_found_,
        _type_hint: typing.Type = None,
        _entry: Entry | None = None,
        _parent: HTree | None = None,
        **kwargs,
    ) -> _T:
        if _type_hint is None:
            _type_hint = self._type_hint_(_key)

        if _type_hint is None:
            return value

        if _parent is None:
            _parent = self

        if isinstance_generic(value, _type_hint):
            pass
        elif issubclass(get_origin(_type_hint), HTree):
            value = _type_hint(value, _entry=_entry, _parent=_parent, **kwargs)

        else:
            if value is not _not_found_:
                pass
            elif _entry is not None:
                value = _entry.get(default_value=default_value)
            else:
                value = default_value

            if value is not _not_found_ and value is not _undefined_ and value is not None:
                value = type_convert(_type_hint, value, **kwargs)

        if isinstance(value, HTreeNode):
            if value._parent is None and _parent is not _not_found_:
                value._parent = _parent

            if len(kwargs) > 0:
                value._metadata.update(kwargs)

            if isinstance(_key, str):
                value._metadata["name"] = _key

            elif isinstance(_key, int):
                value._metadata.setdefault("index", _key)
                if self._metadata.get("name", _not_found_) in (_not_found_, None, "unnamed"):
                    self._metadata["name"] = str(_key)

        return value


def as_htree(*args, **kwargs):
    if len(args) == 0 and len(kwargs) > 0:
        res = Dict(kwargs)
        kwargs = {}
    elif len(args) > 1 and len(kwargs) == 0:
        res = List(list(args))
        args = []
    elif len(args) == 0:
        res = None
    elif isinstance(args[0], HTree):
        res = args[0]
        args = args[1:]
    elif isinstance(args[0], collections.abc.MutableMapping):
        res = Dict(args[0])
        args = args[1:]
    elif isinstance(args[0], collections.abc.MutableSequence):
        res = List(args[0])
        args = args[1:]
    elif len(args) > 1:
        res = List(list(args))
        args = []
    else:
        res = HTree(*args, **kwargs)

    if len(args) + len(kwargs) > 0:
        res._update_(*args, **kwargs)

    return res


class Dict(HTree[_T]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self._cache is _not_found_:
            self._cache = {}

    def children(self) -> typing.Generator[_T, None, None]:
        yield from self.values()

    def __iter__(self) -> typing.Generator[str, None, None]:
        """遍历 children"""
        yield from self.keys()

    def items(self) -> typing.Generator[typing.Tuple[str, _T], None, None]:
        yield from self.for_each()

    def keys(self) -> typing.Generator[str, None, None]:
        for k, v in self.for_each():
            yield k

    def values(self) -> typing.Generator[_T, None, None]:
        for k, v in self.for_each():
            yield v


collections.abc.MutableMapping.register(Dict)


class List(HTree[_T]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self._cache is _not_found_:
            self._cache = []
        elif not isinstance(self._cache, list):
            self._cache = [self._cache]

    def children(self) -> typing.Generator[_T, None, None]:
        for k, v in self.for_each():
            yield v

    def __iter__(self) -> typing.Generator[_T, None, None]:
        for k, v in self.for_each():
            yield v

    def append(self, value):
        return self._insert_(value, Path.tags.append)

    def extend(self, value):
        return self._insert_(value, Path.tags.extend)

    def __iadd__(self, other: list) -> typing.Type[List[_T]]:
        return self.extend(other)


collections.abc.MutableSequence.register(List)



"""
定义一个property, 要求其所在的class必须有一个_as_child方法，用于将其转换为type_hint 指定的类型。
    ```python
        class Foo(Dict):
            pass

        class Doo(Dict):

            f0 = SpProperty(type_hint=Foo)      # 优先级最高, 不兼容IDE的类型提示

            f1: Foo = SpProperty()              # 推荐，可以兼容IDE的类型提示

            ######################################################
            @sp_property
            def f3(self) -> Foo:                 # 用于定义f3的getter操作，与@property.getter类似
                'This is  f3!'
                return self.get("f3", {})

            @f3.setter
            def f3(self,value)->None:            # 功能与@property.setter  类似, NOT IMPLEMENTED YET!!
                self._entry.put("f3",value)

            @f3.deleter
            def f3(self)->None:                  # 功能与@property.deleter 类似, NOT IMPLEMENTED YET!!
                self._entry.child("f3").erase()
            ######################################################
                                                 # 完整版本
            def get_f4(self,default={})->Foo:
                return self.get("f4", default)

            def set_f4(self,value)->None:
                return self.set("f4", value)

            def del_f4(self,value)->None:
                return self.set("f4", value)

            f4 = sp_property(get_f4,set_f4,del_f4,"I'm f4",type_hint=Foo)
        ```

"""

from __future__ import annotations
import pprint
import inspect
import typing
import collections.abc
from copy import deepcopy, copy
from _thread import RLock
from enum import Enum
from typing_extensions import Self

from .Entry import Entry
from .AoS import AoS
from .HTree import HTree, Dict, List, HTreeNode
from .Path import update_tree, merge_tree, Path
from .Expression import Expression
from ..utils.envs import SP_DEBUG
from ..utils.logger import logger, deprecated
from ..utils.tags import _not_found_, _undefined_


def _copy(obj, *args, **kwargs):
    if isinstance(obj, dict):
        return {k: _copy(v, *args, **kwargs) for k, v in obj.items()}

    elif isinstance(obj, list):
        return [_copy(v, *args, **kwargs) for k, v in obj]

    elif isinstance(obj, SpTree):
        cache = {}

        for k, attr in inspect.getmembers(obj.__class__, lambda c: isinstance(c, SpProperty)):
            if attr.getter is None and attr.alias is None:
                try:
                    value = getattr(obj, k, _not_found_)
                except Exception as error:
                    value = _not_found_

                if value is not _not_found_:
                    try:
                        cache[k] = _copy(value, *args, **kwargs)
                    except Exception as error:
                        logger.exception(f"{obj} {value}")
                        raise error

        return cache

    elif isinstance(obj, HTreeNode):
        return obj.fetch(*args, **kwargs)

    else:
        return deepcopy(obj)


class SpTree(Dict[HTreeNode]):
    """支持 sp_property 的 Dict"""

    def __serialize__(self, dumper: typing.Callable[...] | bool = True) -> typing.Dict[str, typing.Any]:
        data = {}
        for k, prop in inspect.getmembers(self.__class__, lambda c: is_sp_property(c)):
            if prop.getter is not None:
                continue
            value = getattr(self, k, _not_found_)
            if value is _not_found_:
                continue
            # elif isinstance(value, Expression):
            #     value = value.__array__()

            data[k] = value

        return super()._do_serialize(data, dumper)

    def fetch(self, *args, exclude=[], **kwargs) -> Self:
        if len(args) + len(kwargs) == 0:  # FIXME: 在 flush 的时候会有问题，需要 debug
            return super().fetch()

        cache = {}

        for k, attr in inspect.getmembers(self.__class__, lambda c: isinstance(c, SpProperty)):
            if k in exclude:
                continue
            if attr.getter is None and attr.alias is None:
                value = getattr(self, k, _not_found_)
                if value is _not_found_:
                    continue
                elif isinstance(value, HTreeNode):
                    cache[k] = value.fetch(*args, **kwargs)
                else:
                    cache[k] = HTreeNode._do_fetch(value, *args, **kwargs)

        return self.__duplicate__(cache, _parent=None)


class PropertyTree(SpTree):
    def __getattr__(self, key: str, *args, **kwargs) -> PropertyTree | AoS:
        if key.startswith("_"):
            return super().__getattribute__(key)
        else:
            res = Path._do_find(self._cache, [key], *args, **kwargs)
            if isinstance(res, dict):
                return PropertyTree(res, _parent=self)
            elif isinstance(res, list) and (len(res) == 0 or isinstance(res[0], (dict, HTree))):
                return AoS[PropertyTree](res, _parent=self)
            elif res is _not_found_:
                return self.__missing__(key)
            else:
                return res

    def __missing__(self, key) -> typing.Any:
        return _not_found_

    def _type_hint_(self, *args, **kwargs):
        return PropertyTree

    def dump(self, entry: Entry | None = None, force=False, quiet=True) -> Entry:
        if entry is None:
            return deepcopy(self._cache)
        else:
            entry.update(self._cache)
            return entry


_T = typing.TypeVar("_T")
_TR = typing.TypeVar("_TR")

"""
    用于为 SpPropertyClass 类（及其子类）定义一个property, 并确保其类型为type_hint 指定的类型。

    例如：
    ``` python
        class Foo(SpPropertyClass):
            # 方法一
            @sp_property
            def a(self) -> float: return 128

            # 方法二
            @sp_property(coordinate1="../psi")
            def dphi_dpsi(self) -> Profile[float]: return self.a*2

            # 方法三
            phi: Profile[float] = sp_property(coordinate1="../psi")

    ```
    方法二、三中参数 coordinate1="../psi"，会在构建 Profile时传递给构造函数  Profile.__init__。

    方法三 会在创建class 是调用 __set_name__,
            会在读写property phi 时调用 __set__,__get__ 方法，
            从Node的_cache或entry获得名为 'phi' 的值，将其转换为 type_hint 指定的类型 Profile[float]。
    """


class SpProperty:
    def __init__(
        self,
        getter: typing.Callable[[typing.Any], typing.Any] = None,
        setter=None,
        deleter=None,
        default_value: typing.Any = _not_found_,
        alias=None,
        type_hint: typing.Type = None,
        doc: str = None,
        strict: bool = True,
        **kwargs,
    ):
        """
        Parameters
        ----------
        getter : typing.Callable[[typing.Any], typing.Any]
            用于定义属性的getter操作，与@property.getter类似
        setter : typing.Callable[[typing.Any, typing.Any], None]
            用于定义属性的setter操作，与@property.setter类似
        deleter : typing.Callable[[typing.Any], None]
            用于定义属性的deleter操作，与@property.deleter类似
        type_hint : typing.Type
            用于指定属性的类型
        doc : typing.Optional[str]
            用于指定属性的文档字符串
        strict : bool
            用于指定是否严格检查属性的值是否已经被赋值
        metadata : typing.Any
            用于传递给构建  Node.__init__ 的参数


        """

        self.lock = RLock()

        self.getter = getter
        self.setter = setter
        self.deleter = deleter
        self.default_value = default_value
        self.alias = alias
        self.doc = doc or ""
        self.property_name: str = None
        self.strict = strict
        self.metadata = kwargs
        self.type_hint = type_hint

    def __call__(self, func: typing.Callable[..., _TR]) -> _TR:
        """用于定义属性的getter操作，与@property.getter类似"""
        if self.getter is not None:
            raise RuntimeError(f"Can not reset getter!")
        self.getter = func
        return self

    def __set_name__(self, owner_cls, name: str):
        # TODO：
        #    若 owner 是继承自具有属性name的父类，则默认延用父类sp_property的设置
        self.property_name = name

        tp = None
        if callable(self.getter):
            tp = typing.get_type_hints(self.getter).get("return", None)

        if tp is None:
            try:
                tp = typing.get_type_hints(owner_cls).get(name, None)
            except Exception as error:
                logger.exception(owner_cls)
                raise error

        if tp is not None:
            self.type_hint = tp

        if self.getter is not None:
            self.doc += self.getter.__doc__ or ""

        for base_cls in owner_cls.__bases__:
            prop = getattr(base_cls, name, _not_found_)
            if isinstance(prop, SpProperty):
                if self.default_value is _not_found_:
                    self.default_value = prop.default_value

                if self.alias is None:
                    self.alias = prop.alias

                self.doc += prop.doc

                self.metadata = update_tree(deepcopy(prop.metadata), self.metadata)

        self.metadata["name"] = name

        if self.doc == "":
            self.doc = f"{owner_cls.__name__}.{self.property_name}"

    def __set__(self, instance: SpTree, value: typing.Any) -> None:
        assert instance is not None

        if self.alias is not None:
            logger.warning(f"set proptery alias {self.property_name} => {self.alias}!")

        if self.property_name is None:
            logger.error("Can not use sp_property instance without calling __set_name__ on it.")

        with self.lock:
            instance._update_(self.property_name, value, _setter=self.setter)

    def __get__(self, instance: SpTree, owner_cls=None) -> _T:
        if instance is None:
            # 当调用 getter(cls, <name>) 时执行
            return self
        elif not isinstance(instance, SpTree):
            raise TypeError(f"Class '{instance.__class__.__name__}' must be a subclass of 'SpTree'.")

        with self.lock:
            if self.alias is not None:
                try:
                    value = instance._find_(
                        self.property_name,  # property_name 必然是 identifier
                        _type_hint=self.type_hint,
                        _getter=self.getter,
                        default_value=_undefined_,
                        **self.metadata,
                    )
                except KeyError as error:
                    value = _not_found_

                if value is _not_found_ or value is _undefined_:
                    # alias 不改变 _parent
                    value = instance.get(
                        self.alias,  # alias 可以是路径
                        _type_hint=self.type_hint,
                        default_value=deepcopy(self.default_value),
                        _parent=_not_found_,
                    )
            else:
                value = instance._find_(
                    self.property_name,
                    _type_hint=self.type_hint,
                    _getter=self.getter,
                    default_value=deepcopy(self.default_value),
                    **self.metadata,
                )

            if self.strict and value is _undefined_:
                raise AttributeError(
                    f"The value of property '{owner_cls.__name__ if owner_cls is not None else 'none'}.{self.property_name}' is not assigned!"
                )

        return value

    def __delete__(self, instance: SpTree) -> None:
        with self.lock:
            instance._remove_(self.property_name, deleter=self.deleter)


def sp_property(getter: typing.Callable[..., _T] | None = None, **kwargs) -> _T:
    if getter is None:
        return SpProperty(**kwargs)
    else:
        return SpProperty(getter=getter, **kwargs)


def is_sp_property(obj) -> bool:
    return isinstance(obj, SpProperty)


def _process_sptree(cls, **kwargs) -> typing.Type[SpTree]:
    if not inspect.isclass(cls):
        raise TypeError(f"Not a class {cls}")

    type_hints = typing.get_type_hints(cls)

    if not issubclass(cls, HTree):
        n_cls = type(cls.__name__, (cls, SpTree), {})
        n_cls.__module__ = cls.__module__
    else:
        n_cls = cls

    for _name, _type_hint in type_hints.items():
        prop = getattr(cls, _name, _not_found_)

        if isinstance(prop, property):
            continue
        elif isinstance(prop, SpProperty):
            if not (_name in cls.__dict__ and n_cls is cls):
                prop = SpProperty(getter=prop.getter, setter=prop.setter, deleter=prop.deleter, **prop.metadata)
        else:
            prop = SpProperty(default_value=prop)

        prop.type_hint = _type_hint

        setattr(n_cls, _name, prop)

        prop.__set_name__(n_cls, _name)

    setattr(n_cls, "_metadata", merge_tree(getattr(cls, "_metadata", {}), kwargs))

    return n_cls


def sp_tree(cls: _T = None, /, **kwargs) -> _T:
    # 装饰器，将一个类转换为 SpTree 类

    def wrap(_cls, _kwargs=kwargs):
        return _process_sptree(_cls, **_kwargs)

    if cls is None:
        return wrap
    else:
        return wrap(cls)



import unittest
from copy import deepcopy

from spdm.core.Path import Path
from spdm.utils.logger import logger
from spdm.utils.tags import _not_found_
from spdm.utils.typing import get_origin, get_args, isinstance_generic
import typing


class TestTyping(unittest.TestCase):

    def test_get_origin(self):
        _T = typing.TypeVar("_T")

        class Foo(typing.Generic[_T]):
            pass

        foo = Foo[int]()

        self.assertEqual(get_origin(int), int)
        self.assertEqual(get_origin(Foo), Foo)
        self.assertEqual(get_origin(Foo[int]), Foo)

        class Goo(Foo[_T]):
            pass

        self.assertEqual(get_origin(Goo), Goo)
        self.assertEqual(get_origin(Goo[int]), Goo)

        class Doo(Foo[float]):
            pass

        self.assertEqual(get_origin(Doo), Doo)

        self.assertEqual(get_origin(1.234), float)
        self.assertEqual(get_origin(Foo()), Foo)
        self.assertEqual(get_origin(Foo[int]()), Foo)
        self.assertEqual(get_origin(Goo[int]()), Goo)
        self.assertEqual(get_origin(Doo()), Doo)

    def test_get_args(self):
        _T = typing.TypeVar("_T")

        class Foo(typing.Generic[_T]):
            pass

        self.assertEqual(get_args(int), tuple())
        self.assertEqual(get_args(Foo), tuple())
        self.assertEqual(get_args(Foo[int]), (int,))

        class Goo(Foo[_T]):
            pass

        self.assertEqual(get_args(Goo), tuple())
        self.assertEqual(get_args(Goo[int])[0], int)

        class Doo(Foo[float]):
            pass

        self.assertEqual(get_args(Doo), (float,))

    def test_isinstance_generic(self):
        _T = typing.TypeVar("_T")

        class Foo(list, typing.Generic[_T]):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
            pass

        foo = Foo[int]()

        self.assertTrue(isinstance_generic(foo, Foo[int]))
        self.assertTrue(isinstance_generic(foo, list))
        self.assertTrue(isinstance_generic(foo, Foo))

        self.assertFalse(isinstance_generic(foo, Foo[float]))
        self.assertFalse(isinstance_generic(foo, dict))

        goo=Foo()
        self.assertTrue(isinstance_generic(goo, Foo))
        self.assertTrue(isinstance_generic(goo, Foo[float]))

if __name__ == '__main__':
    unittest.main()



import tempfile
import unittest

import numpy as np
from spdm.utils.logger import logger
from spdm.data.File import File


class TestFile(unittest.TestCase):
    data = {
        "a": [
            "hello world {name}!",
            "hello world2 {name}!",
        ],
        "b": [1.0, 2, 3, 4],
        "c": "I'm {age}!",
        "d": {
            "e": "{name} is {age}",
            "f": "{address}",
            "g": {"a": 1, "b": 2}
        },
        "h": np.random.random([10, 10])
    }

    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory(prefix="spdm_")
        return super().setUp()

    def tearDown(self) -> None:
        self.temp_dir.cleanup()
        self.temp_dir = None
        return super().tearDown()

    def test_xml(self):
        device_desc = File("/home/salmon/workspace/fytok_data/mapping/ITER/imas/3/static/config.xml",
                           format="XML").read()
        # logger.debug(device_desc.get({"wall", "pf_active", "tf", "magnetics"}).dump())
        # {"wall", "pf_active", "tf", "magnetics"}
        logger.debug(device_desc.child("wall/description_2d/limiter").dump())


if __name__ == '__main__':
    unittest.main()



from .Solid import Solid
from .Plane import Plane
from .Point import Point
from .Line import Line


class Sweep:
    def __init__(self, shape: Plane | Line, axis: Point | Line, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._section = shape
        self._axis = axis

    @property
    def cross_section(self) -> Plane | Line:
        return self._section

    @property
    def axis(self) -> Point | Line:
        return self._axis


class SweepSurface(Solid):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class SweepSolid(Solid):
    Boundary = SweepSurface

    def __init__(self, shape: Plane, axis:  Line, *args, **kwargs) -> None:
        super().__init__(shape, axis, *args, **kwargs)
        self._boundary = self.__class__.Boundary(shape.boundary, axis, *args)



from spdm.data import file
import pprint
from ..util.logger import logger

if __name__ == "__main__":
    d = File("/home/salmon/workspace/fytok/external/SpDB/examples/data/g063982.04800",
                schema="geqdsk")
    print(d._data)
    # io.save(d, "test.json")
    # io.save(d, "test.yaml")
    # # io.save(d, "test.nml")
    # db = io.connect("local://", schema="json")
    # # idx = db.save(d)
    # db.load(idx)



import abc
import typing

import numpy as np
from ..utils.typing import ArrayType
from .GeoObject import GeoObject
from .Line import Line, Segment
from .Plane import Plane
from .Point import Point
from .Polygon import Polygon
from .PointSet import PointSet


class Polyhedron(PointSet):

    def __init__(self, *args,  **kwargs) -> None:
        super().__init__(*args, rank=3, **kwargs)

    @property
    def is_convex(self) -> bool: return True

    @property
    def edges(self) -> typing.Generator[Segment, None, None]:
        raise NotImplementedError()

    @property
    def faces(self) -> typing.Generator[Polygon, None, None]:
        raise NotImplementedError()

    @property
    def boundary(self) -> typing.Generator[Polygon, None, None]:
        yield from self.faces



import collections
import collections.abc
import typing

from ..utils.sp_export import sp_find_module
from ..utils.uri_utils import URITuple, uri_split
from .Document import Document
from .Entry import Entry

InsertOneResult = collections.namedtuple("InsertOneResult", "inserted_id success")
InsertManyResult = collections.namedtuple("InsertManyResult", "inserted_ids success")
UpdateResult = collections.namedtuple("UpdateResult", "inserted_id success")
DeleteResult = collections.namedtuple("DeleteResult", "deleted_id success")


class Collection(Document):
    ''' Collection of documents
    '''
    _registry = {}

    @classmethod
    def _plugin_guess_name(cls, path, *args,  **kwargs) -> typing.List[str]:
        n_cls_name = None

        if "protocol" in kwargs:
            protocol = kwargs.get("protocol")
            n_cls_name = protocol
        elif isinstance(path, collections.abc.Mapping):
            n_cls_name = path.get("$class", None)
        elif isinstance(path, (str, URITuple)):
            uri = uri_split(path)
            n_cls_name = uri.protocol

        return [f"spdm.plugins.data.Plugin{n_cls_name}#{n_cls_name}Collection"]

    def __new__(cls,  *args, **kwargs):
        if cls is not Collection:
            return object.__new__(cls)
        else:
            return super().__new__(cls, *args, **kwargs)

    def __init__(self, uri, *args,  mapper=None,   **kwargs):
        super().__init__(uri, *args, **kwargs)
        self._mapper = mapper

    @property
    def mapper(self):
        return self._mapper

    def guess_id(self, predicate, *args, fragment: int = None, **kwargs) -> int:
        if isinstance(predicate, int):
            return predicate
        elif isinstance(predicate, str):
            return int(predicate)
        elif fragment is not None:
            return int(fragment) if isinstance(fragment, str) else fragment
        else:
            raise NotImplementedError(f"predicate={predicate} fragment={fragment}")

    @property
    def next_id(self):
        raise NotImplementedError()

    def _mapping(self, entry: Entry) -> Entry:
        return self.mapper.map(entry) if self.mapper is not None else entry

    def create_one(self, *args, **kwargs):
        return self.insert_one(*args, mode="x", **kwargs)

    def create_many(self, docs: typing.List[typing.Any], *args, **kwargs):
        return [self.create_one(doc, *args, mode="x", **kwargs) for doc in docs]

    def create_doc(self, docs, *args, **kwargs):
        if isinstance(docs, collections.abc.Sequence):
            return self.create_many(docs, *args, **kwargs)
        else:
            return self.create_one(docs, *args, **kwargs)

    def insert_one(self, doc, * args,  **kwargs) -> InsertOneResult:
        raise NotImplementedError()

    def insert_many(self, docs: typing. List[typing.Any], *args, **kwargs) -> InsertManyResult:
        return [self.insert_one(doc, *args, **kwargs) for doc in docs]

    def insert(self, docs, *args, **kwargs):
        if isinstance(docs, collections.abc.Sequence):
            return self.insert_many(docs, *args, **kwargs)
        else:
            return self.insert_one(docs, *args, **kwargs)

    def find_one(self, *args, **kwargs) -> Entry:
        raise NotImplementedError()

    def find_many(self, *args, **kwargs) -> typing.List[Entry]:
        raise NotImplementedError()

    def find(self, predicate, projection=None, only_one=False, **kwargs) -> typing.Union[Entry,  typing.List[Entry]]:
        # if not isinstance(predicate, str) and isinstance(predicate, collections.abc.Sequence):
        if not only_one:
            return self.find_many(predicate, projection,  **kwargs)
        else:
            return self.find_one(predicate, projection,  **kwargs)

    def replace_one(self, predicate, replacement,  *args, **kwargs) -> UpdateResult:
        raise NotImplementedError()

    def update_one(self, predicate, update,  *args, **kwargs) -> UpdateResult:
        raise NotImplementedError()

    def update_many(self, predicate, updates: list,  *args, **kwargs) -> UpdateResult:
        return [self.update_one(predicate, update, *args, **kwargs) for update in updates]

    def delete_one(self, predicate,  *args, **kwargs) -> DeleteResult:
        raise NotImplementedError()

    def delete_many(self, predicate, *args, **kwargs) -> DeleteResult:
        raise NotImplementedError()

    def count(self, predicate=None, *args, **kwargs) -> int:
        raise NotImplementedError()

    ######################################################################
    # TODO(salmon, 2019.07.01) support index

    def create_indexes(self, indexes: typing.List[str], session=None, **kwargs):
        raise NotImplementedError()

    def create_index(self, keys: typing.List[str], session=None, **kwargs):
        raise NotImplementedError()

    def ensure_index(self, key_or_list, cache_for=300, **kwargs):
        raise NotImplementedError()

    def drop_indexes(self, session=None, **kwargs):
        raise NotImplementedError()

    def drop_index(self, index_or_name, session=None, **kwargs):
        raise NotImplementedError()

    def reindex(self, session=None, **kwargs):
        raise NotImplementedError()

    def list_indexes(self, session=None):
        raise NotImplementedError()


def open_collection(uri: typing.Union[str, URITuple], *args, schema=None, ** kwargs) -> Collection:
    url_ = uri_split(url)
    if url_.protocol is None:
        url_.scheme = "localdb"

    if source_schema is None and url_.protocol != "":
        source_schema = url_.protocol

    mapper = create_mapper(url_.protocol,  schema)

    if url_.protocol == "localdb":
        db = FileCollection(uri, *args, mapper=mapper, **kwargs)
    else:
        db = Collection(uri, *args, mapper=mapper, **kwargs)
    return db


open_db = open_collection



import collections
import inspect

from .logger import logger


class Signature(inspect.Signature):

    @staticmethod
    def _create_annotation(s):
        if type(s) is str:
            return getattr(__builtins__, s, {"type": s})
        elif isinstance(s, type):
            return s
        elif s is None:
            return inspect.Signature.empty
        else:
            # elif isinstance(s, collections.abc.Mapping):
            return s
        # else:
        #     raise TypeError(f"Illegal annotation type! {s}")

    @staticmethod
    def _create_parameter(p, default_kind="KEYWORD_ONLY"):
        if isinstance(p, inspect.Parameter):
            return p
        elif isinstance(p, collections.abc.Mapping):
            return inspect.Parameter(
                p.get("name"),
                inspect._ParameterKind[p.get("kind", default_kind)].value,
                default=p.get("default", inspect.Parameter.empty),
                annotation=Signature._create_annotation(
                    p.get("dtype", p.get("annotation", None)))
            )

    def __init__(self, parameters=None, *, return_annotation=inspect.Signature.empty, default_kind="KEYWORD_ONLY"):

        if type(parameters) is not str and isinstance(parameters, collections.abc.Sequence):
            parameters = [Signature._create_parameter(
                p, default_kind=default_kind) for p in parameters]
        elif isinstance(parameters, collections.abc.Mapping):
            parameters = [Signature._create_parameter(
                p, default_kind=default_kind) for p in parameters.values()]

        super().__init__(parameters=parameters,
                         return_annotation=Signature._create_annotation(return_annotation))

    @classmethod
    def create(cls,  parameters=None, return_annotation=None, **kwargs):
        # assert(type(parameters) is not str and isinstance(
        #     parameters, collections.abc.Sequence))
        # parameters = [p for p in parameters if p.get("direction", "IN") != "OUT"]

        if return_annotation is None:
            pass
        elif isinstance(return_annotation, collections.abc.Sequence) and len(return_annotation) == 1:
            return_annotation = return_annotation[0]

        return cls(parameters, return_annotation=return_annotation, **kwargs)

    @classmethod
    def from_callable(cls, func, *args, **kwargs):
        try:
            sig = super(Signature, cls).from_callable(func)
        except:
            if hasattr(func, "__call__"):
                sig = super(Signature, cls).from_callable(func.__call__)
            else:
                raise TypeError(f"Can not get signature from {type(func)}!")
        return sig




import os

from spdm.core.AoS import AoS
from spdm.core.Expression import Expression, Variable
from spdm.core.Field import Field
from spdm.core.Function import Function, function_like
from spdm.core.HTree import List
from spdm.core.sp_property import sp_property
from spdm.core.TimeSeries import TimeSeriesAoS
from spdm.geometry.Curve import Curve
from spdm.geometry.GeoObject import GeoObject, GeoObjectSet
from spdm.geometry.Point import Point
from spdm.mesh.Mesh import Mesh
from spdm.mesh.mesh_curvilinear import CurvilinearMesh
from spdm.numlib.contours import _find_contours
from spdm.numlib.optimize import minimize_filter
from spdm.utils.constants import *
from spdm.utils.tags import _not_found_
from spdm.utils.tree_utils import merge_tree_recursive
from spdm.utils.typing import (ArrayLike, ArrayType, NumericType, array_type,
                               scalar_type)
from spdm.view.View import draw_profiles

WORKSPACE = "/home/salmon/workspace"

os.environ["SP_DATA_MAPPING_PATH"] = f"{WORKSPACE}/fytok_data/mapping"

input_path = f"{WORKSPACE}/gacode/neo/tools/input/profile_data"

output_path = f"{WORKSPACE}/output"

if __name__ == "__main__":

    _X = Variable(0, "R")

    x = np.linspace(0, scipy.constants.pi*2, 100)

    draw_profiles(
        [
            ([
                (np.sin(_X), {"label": "sin(x)"}),
                (np.cos(_X), {"label": "cos(x)"})
            ], {"y_label": r"$\chi_{e}$", }),
            ([
                (3*_X, {"label": r"$3\times x$"}),
                (4*_X, {"label": r"$4\times x$"})
            ], {"y_label": r"$\chi_{e}$", }),
        ],
        x_value=x,
        x_label="x",
        title="Demo draw profiles",
        styles={"fontsize": 16},
        output=f"{output_path}/demo_draw_profiles.svg"
    )



from __future__ import annotations

import typing
import numpy as np
from spdm.utils.logger import logger

from ..utils.typing import ArrayType, NumericType
from .GeoObject import GeoObject


class Point(GeoObject):
    """Point
    点，零维几何体
    """

    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 1:
            if isinstance(args[0], (list, tuple)):
                args = args[0]
            elif isinstance(args[0], dict):
                args = [*args[0].values()]
        super().__init__(*args, rank=0, ndims=len(args), **kwargs)
        self._coord = np.array(args)

    def __iter__(self) -> typing.Generator[float, None, None]:
        yield from self._coord

    @property
    def r(self) -> float:
        """just a walkaround for spherical coordinate system"""
        return self._coord[0]

    @property
    def z(self) -> float:
        """just a walkaround for spherical coordinate system"""
        return self._coord[1]

    @property
    def x(self) -> float:
        return self._coord[0]

    @property
    def y(self) -> float:
        return self._coord[1]

    @property
    def measure(self) -> float:
        return 0

    @property
    def points(self):
        return self._coord



import sys
sys.path.append("/home/salmon/workspace/SpDev/SpCommon")
sys.path.append("/home/salmon/workspace/SpDev/SpDB")


import matplotlib.pyplot as plt
import numpy as np
from spdm.data import Document
from ..util.logger import logger



# import os

# os.environ['east_path'] = 'mds.ipp.ac.cn::/pcs_east'


if __name__ == '__main__':
    entry = Document([
        "/home/salmon/workspace/SpDev/SpDB/mapping/EAST/imas/3/static/config.xml",
        "/home/salmon/workspace/SpDev/SpDB/mapping/EAST/imas/3/dynamic/config.xml"
    ], format_type="XML").entry

    fg = plt.figure()

    plt.gca().add_patch(plt.Polygon(np.array([entry.wall.description_2d.vessel.annular.outline_outer.r.__value__(),
                                              entry.wall.description_2d.vessel.annular.outline_outer.z.__value__()]).transpose([1, 0]),
                                    fill=False, closed=True))

    plt.gca().add_patch(plt.Polygon(np.array([entry.wall.description_2d.vessel.annular.outline_inner.r.__value__(),
                                              entry.wall.description_2d.vessel.annular.outline_inner.z.__value__()]).transpose([1, 0]),
                                    fill=False, closed=True))

    plt.gca().add_patch(plt.Polygon(np.array([entry.wall.description_2d[0].limiter.unit[0].outline.r.__value__(),
                                              entry.wall.description_2d[0].limiter.unit[0].outline.z.__value__()]).transpose([1, 0]),
                                    fill=False, closed=True))

    for coil in entry.pf_active.coil:
        rect = coil.element[0].geometry.rectangle.__value__()
        logger.debug(rect)
        plt.gca().add_patch(plt.Rectangle((rect.r-rect.width/2.0, rect.z-rect.height/2.0), rect.width, rect.height, fill=False))

    # logger.debug([time_slice.profiles_2d[0].boundary.type.__value__()
    #               for time_slice in entry.equilibrium.time_slice[1:10]])

    # plt.contour(
    #     entry.equilibrium.time_slice[1].profiles_2d.Mesh.dim1.__value__(),
    #     entry.equilibrium.time_slice[1].profiles_2d.Mesh.dim2.__value__(),
    #     entry.equilibrium.time_slice[1].profiles_2d.psi.__value__(),
    #     levels =30,linewidths=0.4
    #     )
    plt.axis('scaled')

    # for time_slice in entry.equilibrium.time_slice[:]:
    #     logger.debug(time_slice.ids_properties.homogeneous_time.__value__())
    plt.savefig("imas_east.png")



import sys
import numpy as np
import timeit

sys.path.append("/home/salmon/workspace/SpDev/SpDB")

if __name__ == "__main__":
    from ..util.logger import logger
    from spdm.core.Profile import Profile, Profiles

    npoints = 100

    x_axis = np.linspace(0, 1.0, npoints)

    x, y, z = np.random.rand(3, npoints)

    X = x.view(Profile)

    X._axis = x_axis
    Y = y.view(Profile)
    Y._axis = x_axis
    Z = z.view(Profile)
    Z._axis = x_axis

    r = x + (y**2 + (z*x + 1)*3)
    R = X + (Y**2 + (Z*X + 1)*3)

    logger.debug(R.description)

    logger.debug(all(r == R))

    time0 = timeit.timeit(lambda: x + (y**2 + (z*x + 1)*3), number=100)
    time1 = timeit.timeit(lambda: R.evaluate(), number=100)

    logger.debug((time0, time1))
    
    logger.debug(R.derivative)